import numpy as np
import math
import copy
from typing import Callable, Tuple, Optional, Union
from scipy.integrate import quad

from pygdsdesign.polygons import Rectangle
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.transmission_lines.transmission_line import TransmissionLine
from pygdsdesign.typing_local import Coordinate
from pygdsdesign.shapes import butterworth_filter
from pygdsdesign.functions import distance

class CPWPolar(TransmissionLine):

    def __init__(
        self,
        width: float,
        gap: float,
        angle: float,
        layer: int = 0,
        datatype: int = 0,
        name: str = "",
        color: str = "",
        ref: Optional[Coordinate] = None,
    ) -> None:
        """
        Coplanar allows to easily draw a continuous coplanar waveguide.

        Parameters
        ----------
        width : float
            Width of the central line in um
            This width can be modified latter along the strip or smoothly
            by using tappered functions.
        gap : float
            Width of the central line in um
            This width can be modified latter along the strip or smoothly
            by using tappered functions.
        angle: float
            Orientation of the microstrip in radian.
            This angle can be modified latter with the add_turn function.
            A value of 0 corresponds to the direction left to right.
        layer : int
            Layer number of the coplanar. Default to 0
        datatype : int
            Datatype number of the coplanar. Default to 0
        name: str
            Name of the complanar
        color: str
            Color of the complanar
        """

        TransmissionLine.__init__(self, layer=layer,
                                        datatype=datatype,
                                        name=name,
                                        color=color,
                                        ref=ref)

        self._w = width
        self._angle = angle
        self._s = gap
        self._bounding_polygon = PolygonSet()


    @property
    def width(self):
        return self._w


    @width.setter
    def width(self, width:float):
        self._w = width


    @property
    def gap(self):
        return self._s


    @gap.setter
    def gap(self, gap:float):
        self._s = gap


    @property
    def bounding_polygon(self):
        return self._bounding_polygon

    ###########################################################################
    #
    #                   Add polygons to the existing coplanar waveguide
    #
    ###########################################################################


    def add_line(self, l_len: float) -> PolygonSet:

        """
        Add a piece of linear coplanar in the direction of the angle.

        Parameters
        ----------
        l_len : float
            Length of the strip in um.
        """
        p  = PolygonSet([[(0, -self._w/2.),
                          (0, -self._w/2. - self._s),
                          (l_len, -self._w/2. - self._s),
                          (l_len, -self._w/2.)]],
                         layers=[self._layer],
                         datatypes=[self._datatype],
                         names=[self._name],
                         colors=[self._color])

        p += PolygonSet([[(0, +self._w/2.),
                            (0, +self._w/2. + self._s),
                            (l_len, +self._w/2. + self._s),
                            (l_len, +self._w/2.)]],
                        layers=[self._layer],
                        datatypes=[self._datatype],
                        names=[self._name],
                        colors=[self._color])

        a, b = p.get_bounding_box()
        self._add(p.rotate(self._angle).translate(*self.ref))

        # update bounding polygon

        bp = Rectangle((a[0], a[1]),
                       (b[0], b[1])).rotate(self._angle).translate(*self.ref)

        self.ref = [self.ref[0] + l_len* \
                    np.cos(1*self._angle), self.ref[1] + l_len*np.sin(1*self._angle)]

        self.total_length += abs(l_len)
        self._bounding_polygon += bp

        return self


    ###########################################################################
    #
    #                       Add turn
    #
    ###########################################################################


    def add_turn(self, radius: float,
                       delta_angle: float,
                       nb_points: int=50) -> PolygonSet:
        """
        Add a circulare turn to the strip.

        Parameters
        ----------
        radius : float
            Radius of the arc in um.
        delta_angle : float
            Angle of the turn. a positive value will produces a left turn. A
            A negative value will produces a right turn.
            The angle is relative to the previous angle.
            Hence, a value of pi/2 will produces a 90° left turn,
            relatives to the direction of the last strip.
        nb_point : int (default=50)
            Number of point used in the polygon.
        """

        if delta_angle >= 0:
            start= self._angle - np.pi/2
        else:
            start= self._angle + np.pi/2
        stop= start + delta_angle
        theta=np.linspace(start,stop, nb_points)

        x0 = self.ref[0] + -radius*np.cos(start)
        y0 = self.ref[1] + -radius*np.sin(start)

        x = np.concatenate(((radius+self._w/2.)*np.cos(theta), (radius+self._w/2.+self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((radius+self._w/2.)*np.sin(theta), (radius+self._w/2.+self._s)*np.sin(theta[::-1])))

        p = PolygonSet(polygons=[np.vstack((x+x0, y+y0)).T],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])

        x = np.concatenate(((radius-self._w/2.)*np.cos(theta), (radius-self._w/2.-self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((radius-self._w/2.)*np.sin(theta), (radius-self._w/2.-self._s)*np.sin(theta[::-1])))

        p += PolygonSet(polygons=[np.vstack((x+x0, y+y0)).T],
                        layers=[self._layer],
                        datatypes=[self._datatype],
                        names=[self._name],
                        colors=[self._color])

        # generate bounding polygon
        x = np.concatenate(((radius+self._w/2.+self._s)*np.cos(theta), (radius-self._w/2.-self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((radius+self._w/2.+self._s)*np.sin(theta), (radius-self._w/2.-self._s)*np.sin(theta[::-1])))

        bp = PolygonSet(polygons=[np.vstack((x+x0, y+y0)).T])

        self._angle += delta_angle
        self.ref = [x0+(x[nb_points]+x[nb_points-1])/2,
                    y0+(y[nb_points]+y[nb_points-1])/2]
        self._add(p)
        self._bounding_polygon+=bp
        self.total_length += abs(radius*delta_angle)

        return self



    ###########################################################################
    #
    #                               Tapers
    #
    ###########################################################################


    def add_taper(self, l_len: float,
                        new_width: float,
                        new_gap: float) -> PolygonSet:
        """
        Add linear taper between the current and the new width.

        Parameters
        ----------
        l_len : float
            Length of the taper in um.
        new_width : float
            New width of the microstrip in um.
        new_gap : float
            New gap of the microstip in um.
        """

        p = PolygonSet(polygons=[[(0., self._w/2.)]],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])
        p>(l_len, new_width/2.-self._w/2.)
        p>(0., new_gap)
        p>(-l_len, -new_gap-new_width/2.+self._s+self._w/2.)
        p += copy.copy(p).mirror((0, 0), (1, 0))
        p.rotate(self._angle)

        # generate bounding polygon
        bp = PolygonSet(polygons=[[(0, self._w/2+self._s),
                                  (l_len, new_width/2+new_gap),
                                  (l_len, 0),
                                  (0, 0)]]
                          )
        bp += copy.copy(bp).mirror((0, 0), (1, 0))

        p.translate(*self.ref)
        bp.translate(*self.ref).rotate(self._angle,[self.ref[0],self.ref[1]])

        self.ref = [self.ref[0]+l_len*np.cos(self._angle), self.ref[1]+l_len*np.sin(self._angle)]

        self._add(p)
        self._bounding_polygon+=bp

        self.total_length += abs(l_len)
        self._w = new_width
        self._s = new_gap

        return self

    ###########################################################################
    #
    #                   Generic parametric curve
    #
    ###########################################################################


    def add_parametric_curve(self,
                             f: Callable[..., Tuple[np.ndarray, np.ndarray]],
                             df: Callable[..., Tuple[np.ndarray, np.ndarray]],
                             t: np.ndarray,
                             args: Optional[Tuple[Optional[float], ...]]=None,
                             add_polygon: bool=True,
                             add_length: bool=True) -> Union[PolygonSet, Tuple[PolygonSet, PolygonSet]]:
        """
        Create a coplanar line following the parametric equation f and its
        derivative df along the length t.
        In order to return the curve length correctly, the derivative df of f
        must be correct, its absolute amplitude must be correct.
        The curve is automtically aligned and rotated according to the previous strip angle.
        The next strip's angle will also be changed according to the curve.

        Parameters
        ---------
        f : func
            Function calculating the parametric equation.
            Must be of the type f(t, args) and return a tuple of coordinate
            (x, y).
        df : func
            Function calculating the derivative of the parametric equation.
            Must be of the type df(t, args) and return a tuple of coordinate
            (dx, dy).
        t : np.ndarray
            Array of the length of the parametric curve.
            Also determine the number of point of the total polygon.
            Must not necessarily starts at 0.
        args : variable arguments (default None)
            Argument passed to f and df
        """
        if args is None:
            args = (None, )

        dx1, dy1 = df(t, args)
        n = np.hypot(dx1, dy1)
        dx1, dy1 = dx1/n, dy1/n

        x1, y1 = f(t, args)
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.
        x1, y1 = x1+np.cos(theta1)*(self._w/2.+self._s), y1+np.sin(theta1)*(self._w/2.+self._s)

        dx2, dy2 = df(t[::-1], args)
        n = np.hypot(dx2, dy2)
        dx2, dy2 = dx2/n, dy2/n
        x2, y2 = f(t[::-1], args)
        theta2 = np.angle(dx2+1j*dy2)-np.pi/2.
        x2, y2 = x2+np.cos(theta2)*self._w/2., y2+np.sin(theta2)*self._w/2.

        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p1 = np.vstack((x, y)).T
        # keep the coordinates of the outer trace
        bp_x, bp_y = x1, y1

        dx1, dy1 = df(t, args)
        n = np.hypot(dx1, dy1)
        dx1, dy1 = dx1/n, dy1/n

        x1, y1 = f(t, args)
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.
        x1, y1 = x1+np.cos(theta1)*-(self._w/2.+self._s), y1+np.sin(theta1)*-(self._w/2.+self._s)


        dx2, dy2 = df(t[::-1], args)
        n = np.hypot(dx2, dy2)
        dx2, dy2 = dx2/n, dy2/n
        x2, y2 = f(t[::-1], args)
        theta2 = np.angle(dx2+1j*dy2)-np.pi/2.
        x2, y2 = x2+np.cos(theta2)*-self._w/2., y2+np.sin(theta2)*-self._w/2.

        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p2 = np.vstack((x, y)).T
        # add the coordinates of the outer trace in reversed order
        bp_x, bp_y = np.concatenate((bp_x, np.flip(x1))), np.concatenate((bp_y, np.flip(y1)))
        bp = np.vstack((bp_x, bp_y)).T

        p = PolygonSet(polygons=[p1],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color]).translate(self.ref[0]-x[0]+self._rot(-self._w/2-self._s,0,-theta1[0])[0], self.ref[1]-y[0]+self._rot(-self._w/2-self._s,0,-theta1[0])[1]).rotate(self._angle - theta1[0] - np.pi/2, [self.ref[0], self.ref[1]])\
           + PolygonSet(polygons=[p2],
                        layers=[self._layer],
                        datatypes=[self._datatype],
                        names=[self._name],
                        colors=[self._color]).translate(self.ref[0]-x[0]+self._rot(-self._w/2-self._s,0,-theta1[0])[0], self.ref[1]-y[0]+self._rot(-self._w/2-self._s,0,-theta1[0])[1]).rotate(self._angle - theta1[0] - np.pi/2, [self.ref[0], self.ref[1]])
        # generate bounding_polygon
        bp = PolygonSet(polygons=[bp])
        bp = bp.translate(self.ref[0]-x[0] - (self._w/2 + self._s)*np.cos(theta1[0]), self.ref[1]-y[0] - (self._w/2 + self._s)*np.sin(theta1[0])
                          ).rotate(self._angle-theta1[0]-np.pi/2, [self.ref[0], self.ref[1]])
        self._angle += theta1[-1] - theta1[0]
        # Calculate curve length
        def func(t, args):
            dx1, dy1 = df(t, args)
            return np.hypot(dx1, dy1)

        # Add the length of the parametric curve only if asked (default)
        if add_length:
            self.total_length += quad(func, t[0], t[-1], args=(args,))[0]

        # Add polygon only if asked (default)
        if add_polygon:
            self.ref = [p.polygons[0][int(len(p.polygons[0])/2)][0] -self._w/2*np.cos(self._angle-np.pi/2),
                        p.polygons[0][int(len(p.polygons[0])/2)][1] - self._w/2*np.sin(self._angle-np.pi/2)]
            self._add(p)
            self._bounding_polygon+=bp

            return self
        else:
            return p, bp

    ###########################################################################
    #
    #                             Ends
    #
    ###########################################################################

    def add_end(self, width: float,
                      update_ref: bool=False) -> PolygonSet:
        """
        Add an end to a coplanar waveguide in the perpendicular direction

        Parameters
        ----------
        width : float
            width of the end in um
        """

        r = Rectangle((-self._w/2.-self._s, width),
                      (self._w/2.+self._s, 0),
                          layer=self._layer,
                          datatype=self._datatype,
                          name=self._name,
                          color=self._color)
        a,b = r.get_bounding_box()
        self._add(r.translate(*self.ref).rotate(self._angle-np.pi/2,[self.ref[0],self.ref[1]]))

        # update bounding polygon
        bp = Rectangle((a[0], a[1]),
                        (b[0], b[1])).translate(*self.ref).rotate(self._angle-np.pi/2,[self.ref[0],self.ref[1]])
        self._bounding_polygon+=bp

        if update_ref:
            self.ref = [self.ref[0]+width*np.cos(self._angle), self.ref[1]+width*np.sin(self._angle)]

        return self


    def add_circular_end(self, nb_points: int=50,
                               update_ref: bool=False) -> PolygonSet:
        """
        Add a circular open end to a coplanar waveguide in the given
        orientation.

        Parameters
        ----------
        nb_point : int (default=50)
            Number of point used in the polygon.
        """

        theta = np.linspace(-np.pi/2, np.pi/2, nb_points)
        x = np.concatenate(((self._w/2.)*np.cos(theta), (self._w/2.+self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((self._w/2.)*np.sin(theta), (self._w/2.+self._s)*np.sin(theta[::-1])))
        p = PolygonSet(polygons=[np.vstack((x, y)).T],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])

        # generate bounding polygon
        x = np.array(((self._w/2.+self._s)*np.cos(theta[::-1])))
        y = np.array(((self._w/2.+self._s)*np.sin(theta[::-1])))
        bp = PolygonSet(polygons=[np.vstack((x, y)).T])

        p.rotate(self._angle)
        bp.rotate(self._angle)
        self._add(p.translate(*self.ref))
        self._bounding_polygon+=bp.translate(*self.ref)

        if update_ref:
            added_ref = self._rot(self._w/2 + self._s, 0,self._angle)
            self.ref = [self.ref[0]+added_ref[0], self.ref[1]-added_ref[1]]
        return self


    ###########################################################################
    #
    #                             misc
    #
    ###########################################################################

    def add_butterworth_filter(self,central_conductor_width:float=4,
                       central_conductor_gap:float=0.2,
                       sep_bot_top:float=6,
                       sep_antenna_indutance:float=6,
                       sep_inductance_antenna:float=6,
                       sep_antenna_central:float=6,#
                       nb_l_horizontal:int=1,
                       len_l_horizontal:float=70,
                       len_l_vertical:float=5,
                       l_microstrip_width:float=0.5,#
                       c_arm_length1:list=[18,26,34,34,26,18],
                       c_arm_length2:list=[18,28,38,38,28,18],
                       c_arm_length3:list=[34,26,18],
                       c_central_width:float=2,
                       arm_width:float=2,
                       gap:float=0.2,
                       length:list=[14,32],
                       )-> PolygonSet:
        """
        Return a 5th-order, cauer topology, butterworth filter and its bounding polygons.
        see \pygdsdesign\examples\butterworth_filter_parameters.png for a graphical reprensations of the parameters.

        Args:
            central_conductor_width: width of the central CPW.
                Defaults to 4 um
            central_conductor_gap: gap of the central CPW.
                Defaults to 0.2um
            sep_bot_top: distance added at the start and at the end of the filter. used to separate the filter from other elements.
                Defaults to 6um.
            sep_antenna_indutance: distance between the first capacitance and the first indutance, and between the second inductance and the third capacitance.
                Defaults to 6um.
            sep_inductance_antenna: distance between the first inductance and the second capacitance and between the second capacitance and the second inductance.
                Defaults to 6um.
            sep_antenna_central: distance between the capacitance and the central conductor.
                Defaults to 6um.
            nb_l_horizontal: inductance parameters. number of time the microstrip will go from one side to an other, the first and the last half-length microstrip dont count.
                Defaults to 1.
            len_l_horizontal: inductance parameters. length of the microstrip going from the left (or right) to the right (or left) side.
                Defaults to 70 um.
            len_l_vertical: inductance parameters. distance between two horizontal microstrip.
                Defaults to 5 um.
            l_microstrip_width: inductance parameters. width of the microstrip.
                Defaults to 0.5um.
            c_arm_length1/2/3: capacitance parameters. length of the arm of the first and third capacitance/ the first and the third part of the second capacitance / the second part of the second capacitance.
                Defaults to [18,24,36,24,18] [um].
            c_central_width: capacitance parameters. width of the central conductor, connecting the arms.
                Defaults to 2 um.
            arm_width: capacitance parameters. width of the arms.
                Defaults to 2 um.
            gap: capacitance parameters. distance between the the conductor and the ground plane.
                Defaults to 0.2 um.
            length: capacitance parameters. lengths of the central conductor of the capacitance, connecting the arms.
                Defaults to [14,32]um.
            layer,datatype,name,color: Used for naming the metal layer
                Defaults to 0,0,'',''.
        """

        b_filter,bp= butterworth_filter(central_conductor_width,central_conductor_gap,sep_bot_top,sep_antenna_indutance,sep_inductance_antenna,sep_antenna_central,nb_l_horizontal,len_l_horizontal,len_l_vertical,l_microstrip_width,c_arm_length1,c_arm_length2,c_arm_length3,c_central_width,arm_width,gap,length,
                          layer=self._layer,
                          datatype=self._datatype,
                          name=self._name,
                          color=self._color)

        d=b_filter.get_size()[1]
        self._bounding_polygon += bp.rotate(self._angle-np.pi/2).translate(*self.ref).change_layer(1)
        self._add(b_filter.rotate(self._angle-np.pi/2).translate(*self.ref))
        self.ref = [self.ref[0]+ np.cos(self._angle) * d, self.ref[1]+ np.sin(self._angle) * d]
        return self


    def goto(self, p_destination:PolygonSet)-> PolygonSet:
        """
        Create a link between two CPWs.
        (If someone, want to enhance this function, they may want to start from scratch.)
        Args:
            p_destination: CPWPolar instance of the destination.
        """
        total_half_width= self._w/2 + self._s
        ref2=p_destination.ref
        angle2=(p_destination._angle+np.pi) %(np.sign(p_destination._angle+1e-9)*2*np.pi) #we add 1e-9 because np.sign(0) returns 0, and x % 0 is undefined.
        if angle2 > np.pi:
            angle2-=2*np.pi
        if angle2 < -np.pi:
            angle2+=2*np.pi

        ref1=self.ref

        angle1=self._angle %(np.sign(self._angle+1e-9)*2*np.pi)
        if angle1 > np.pi:
            angle1-=2*np.pi
        if angle1 < -np.pi:
            angle1+=2*np.pi

        x1=(ref1[0]+np.cos(angle1-np.pi/2)*total_half_width)
        y1=(ref1[1]+np.sin(angle1-np.pi/2)*total_half_width)

        x2=(ref1[0]-np.cos(angle1-np.pi/2)*total_half_width)
        y2=(ref1[1]-np.sin(angle1-np.pi/2)*total_half_width)

        x3=ref2[0]+np.cos(angle2-np.pi/2)*total_half_width
        y3=ref2[1]+np.sin(angle2-np.pi/2)*total_half_width

        x4=ref2[0]-np.cos(angle2-np.pi/2)*total_half_width
        y4=ref2[1]-np.sin(angle2-np.pi/2)*total_half_width

        xy=[distance(x1,y1,x3,y3),distance(x1,y1,x4,y4),distance(x2,y2,x3,y3),distance(x2,y2,x4,y4)]
        xy_min=min(xy)
        min_= np.where(xy==xy_min)[0][0]

        if min_==0:
            dx=x3-x1
            dy=y3-y1
        if min_==1:
            dx=x4-x1
            dy=y4-y1
        if min_==2:
            dx=x3-x2
            dy=y3-y2
        if min_==3:
            dx=x4-x2
            dy=y4-y2

        if min_ == 0 or min_ ==3:
            x=math.atan2(dy,dx)%(2*np.pi)

            if x > np.pi:
                x-=2*np.pi
            if x < -np.pi:
                x+=2*np.pi

            turn1 = (x - angle1)

            if turn1 > np.pi:
                turn1-=2*np.pi
            if turn1 < -np.pi:
                turn1+=2*np.pi

            self.add_turn(total_half_width, turn1)
            self.add_line(xy_min)

            turn2=(angle2-x)

            if turn2 > np.pi:
                turn2-=2*np.pi
            if turn2 < -np.pi:
                turn2+=2*np.pi

            self.add_turn(total_half_width, turn2)

        else:
            if min_==2:
                self.add_turn(total_half_width,np.pi/2)
            if min_==1:
                self.add_turn(total_half_width,-np.pi/2)
            self.goto(p_destination)
        return self


import numpy as np
import copy

from pygdsdesign import GdsLibrary, PolygonSet, Rectangle, Text, grid_cover, CPW

# Chip size of 5mm x 7mm
chip_dx = 5000
chip_dy = 7000

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')


# Create empty polygon to which other polygon will be added
tot = PolygonSet()

# Chip mark
p = Rectangle((0,0), (200, 10), layer=0).center()
p += Rectangle((0,0), (10, 200), layer=0).center()
dx_d, dy_d = p.get_size()
tot += p
tot += copy.deepcopy(p).translate(chip_dx, 0)
tot += copy.deepcopy(p).translate(chip_dx, -chip_dy)
tot += copy.copy(p).translate(0., -chip_dy)


# Create design polygon which will contain one feedline and one resonator
p = PolygonSet()

# Feedline
m = CPW(width=500, gap=100, layer=0)
m.add_end(100, 0)
m.add_line(400, 0)
m.add_taper(300, 0, 4, 2)
m.add_line(chip_dx-(500+100)*2., 0)
m.add_taper(300, 0, 500, 100)
m.add_line(400, 0)
m.add_end(100, 0)
p += m
dx_f, dy_f = m.get_size()


# Resonators
res_lengths = [5500, 6500, 7500, 8455]
distance_between_res = 750.
coupling_length = 250
turn_radius = 50
gap_distance = 500
nb_turn = 15
spacing = 30
end_line = 100
ends = ['shunt', 'open_circular', 'open_fresnel', 'open']

# Create design polygon which will contain one feedline and one resonator
pp = PolygonSet()
for i, (res_length, end) in enumerate(zip(res_lengths, ends)):
    m = CPW(width=4, gap=2, layer=0)
    m.add_line(coupling_length, 0)
    m.add_turn(turn_radius, 'lb')
    m.add_line(0, -gap_distance)
    m.add_serpentine(total_length=res_length-coupling_length-gap_distance-turn_radius*np.pi/2.-end_line,
                    nb_turn=nb_turn,
                    spacing=spacing,
                    orientation='bottom',
                    starting='left',
                    )
    m.add_line(0, -end_line)
    if end=='shunt':
        pass
    elif end=='open_circular':
        m.add_circular_end('bottom')
    elif end=='open_fresnel':
        m.add_fresnel_end('bottom')
    elif end=='open':
        m.add_end(0, -m.gap)
    print('Resonator ', i, ' length: ', m.total_length, ' um')
    pp += m.translate(i*distance_between_res, -10.)

dx_r, dy_r = pp.get_size()
p += pp.translate(dx_f/2.-dx_r/2., 0.)

# Copy the feedline and the resonator 4 times
p += copy.copy(p).translate(0, -1500)
p += copy.copy(p).translate(0, -3000)


# Add label on the top right of the chip
t = Text('R0C0', 250)
dx_t, dy_t = t.get_size()
tot += t.translate(-dx_t+chip_dx-250, -dy_t-250)


# Center the design to the chip and the chip to the (0,0) of the cell
tot.center()
tot += p.center()

# Here, we copy the all chip and inverse its polarity
# This will create a polygon of where all the metal is
g = copy.deepcopy(tot).inverse_polarity()
# That polygon is is then cover by a grid of square by using the grid_cover
# function
g = grid_cover(g)

# Add polygons to cell
cell.add(tot)
cell.add(g)

# Save GDS file
lib.write_gds('cpw_resonator_with_grid_gnd.gds')



import numpy as np
import copy
from typing import Callable, Tuple, Optional, Union
from scipy.integrate import quad

from pygdsdesign.polygons import Rectangle
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.transmission_lines.transmission_line import TransmissionLine
from pygdsdesign.typing_local import Coordinate


class MicroStripPolar(TransmissionLine):

    def __init__(self, width: float,
                       angle: float,
                       layer: int=0,
                       datatype: int=0,
                       name: str='',
                       color: str='',
                       ref: Optional[Coordinate] = None,)-> None:
        """
        Microstrip allows to easily draw a continuous microstrip line.

        Parameters
        ----------
        width : float
            Width of the microstrip in um
            This width can be modified latter along the strip or smoothly
            by using tappered functions.
        angle: float
            Orientation of the microstrip in radian.
            This angle can be modified latter with the add_turn function.
            A value of 0 corresponds to the direction left to right.
        layer : int
            Layer number of the coplanar. Default to 0
        datatype : int
            Datatype number of the coplanar. Default to 0
        name: str
            Name of the complanar
        color: str
            Color of the complanar
        """

        TransmissionLine.__init__(self, layer=layer,
                                        datatype=datatype,
                                        name=name,
                                        color=color,
                                        ref=ref)

        self._w = width
        self._angle=angle


    @property
    def width(self):
        return self._w
    def angle(self):
        return self._angle


    @width.setter
    def width(self, width:float):

        self._w = width


    @property
    def param_curve(self):
        return self._param_curve



    ###########################################################################
    #
    #                   Add polygons to the existing microstrip
    #
    ###########################################################################

    def add_line(self, l_len: float) -> PolygonSet:
        """
        Add a piece of linear microstrip direction of the angle.

        Parameters
        ----------
        l_len : float
            Length of the strip in the x direction in um.
        """
        r = Rectangle((self.ref[0], self.ref[1]-self._w/2),
                              (self.ref[0]+l_len, self.ref[1]+self._w/2),
                              layer=self._layer,
                              datatype=self._datatype,
                              name=self._name,
                              color=self._color).rotate(self._angle,[self.ref[0],self.ref[1]])
        self._add2param(self._rot(self.ref[0],self.ref[1]+self._w/2,-self._angle),
                        self._rot(self.ref[0]+l_len, self.ref[1]+self._w/2,-self._angle),
                        [0, abs(l_len)])
        self._add(r)
        self.ref = [self.ref[0] + self._rot(l_len, 0,-1*self._angle)[
            0], self.ref[1] + self._rot(l_len, 0,-1*self._angle)[1]]
        self.total_length += abs(l_len)

        return self


    ###########################################################################
    #
    #                       Add turn
    #
    ###########################################################################


    def add_turn(self, radius: float,
                       delta_angle: float,
                       nb_points: int=50) -> PolygonSet:
        """
        Add a circulare turn to the strip.

        Parameters
        ----------
        radius : float
            Radius of the arc in um.
        delta_angle : float
            Angle of the turn. a positive value will produces a left turn. A
            A negative value will produces a right turn.
            The angle is relative to the previous angle.
            Hence, a value of pi/2 will produces a 90° left turn,
            relatives to the direction of the last strip.
        nb_points : int (default=50)
            Number of point used in the polygon.
        """

        if delta_angle >= 0:
            start= self._angle - np.pi/2
            stop= start + delta_angle
            theta=np.linspace(start,stop, nb_points)
            x0 = self.ref[0] + -radius*np.cos(start)
            y0 = self.ref[1] + -radius*np.sin(start)
            x = np.concatenate(((radius-self._w/2.)*np.cos(theta), (radius + self._w/2.)*np.cos(theta[::-1])))
            y = np.concatenate(((radius-self._w/2.)*np.sin(theta), (radius + self._w/2.)*np.sin(theta[::-1])))
            self.ref = [x0+(x[nb_points]+x[nb_points-1])/2, y0+(y[nb_points]+y[nb_points-1])/2]

        else:
            start= self._angle + np.pi/2
            stop= start + delta_angle
            theta=np.linspace(start,stop, nb_points)
            x0 = self.ref[0] + -radius*np.cos(start)
            y0 = self.ref[1] + -radius*np.sin(start)
            x = np.concatenate(((radius-self._w/2.)*np.cos(theta), (radius + self._w/2.)*np.cos(theta[::-1])))
            y = np.concatenate(((radius-self._w/2.)*np.sin(theta), (radius + self._w/2.)*np.sin(theta[::-1])))
            self.ref = [x0+(x[nb_points]+x[nb_points-1])/2, y0+(y[nb_points]+y[nb_points-1])/2]

        self._angle += delta_angle

        p = np.vstack((x0+x, y0+y)).T

        self._add(PolygonSet(polygons=[p],
                             layers=[self._layer],
                             datatypes=[self._datatype],
                             names=[self._name],
                             colors=[self._color]))

        self.total_length += radius*delta_angle

        self._add2param(x0+radius*np.cos(theta),
                        y0+radius*np.sin(theta),
                        np.linspace(0, 1, nb_points)*radius*np.pi/2)

        return self


    ###########################################################################
    #
    #                               Tapers
    #
    ###########################################################################

    def add_taper(self, l_len: float, new_width: float) -> PolygonSet:

        """
        Add linear taper between the current and the new width.

        Parameters
        ----------
        l_len : float
            Length of the taper in um.
        new_width : float
            New width of the microstrip in um.
        """

        p = [(self.ref[0], self.ref[1]-self._w/2),
             (self.ref[0]+l_len, self.ref[1] - new_width/2.),
             (self.ref[0]+l_len, self.ref[1] + new_width/2.),
             (self.ref[0], self.ref[1]+self._w/2)]

        self._add2param([self.ref[0], self.ref[0]+l_len],
                        [self.ref[1]+self._w/2, self.ref[1]+self._w/2],
                        [0, abs(l_len)])

        self._add(PolygonSet(polygons=[p],
                             layers=[self._layer],
                             datatypes=[self._datatype],
                             names=[self._name],
                             colors=[self._color]).rotate(self._angle,[self.ref[0],self.ref[1]]))
        #self.ref = [self.ref[0]+l_len, self.ref[1]]
        self.ref = [self.ref[0] + self._rot(l_len , 0,-self._angle)[0], self.ref[1] + self._rot(l_len , 0, -self._angle)[1]]
        self.total_length += abs(l_len)

        self._w = new_width

        return self

    ###########################################################################
    #
    #                   Generic parametric curve
    #
    ###########################################################################



    def add_parametric_curve(self, f: Callable[..., Tuple[np.ndarray, np.ndarray]],
                                  df: Callable[..., Tuple[np.ndarray, np.ndarray]],
                                   t: np.ndarray,
                                args: Optional[Tuple[Optional[float], ...]]=None,
                                add_polygon: bool=True,
                                add_length: bool=True) -> PolygonSet:
        """
        Create a microstrip line following the parametric equation f and its
        derivative df along the length t.
        In order to return the curve length correctly, the derivative df of f
        must be correct, its absolute amplitude must be correct.
        The curve is automtically aligned and rotated according to the previous strip angle.
        The next strip's angle will also be changed according to the curve.

        Parameters
        ---------
        f : func
            Function calculating the parametric equation.
            Must be of the type f(t, args) and return a tuple of coordinate
            (x, y).
        df : func
            Function calculating the derivative of the parametric equation.
            Must be of the type df(t, args) and return a tuple of coordinate
            (dx, dy).
        t : np.ndarray
            Array of the length of the parametric curve.
            Also determine the number of point of the total polygon.
            Must not necessarily starts at 0.
        args : variable arguments (default None)
            Argument passed to f and df
        """

        if args is None:
            args = (None, )

        dx1, dy1 = df(t, args)

        n = np.hypot(dx1, dy1)

        dx1, dy1 = dx1/n, dy1/n

        x1_, y1_ = f(t, args)
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.

        x1, y1 = x1_+np.cos(theta1)*self._w/2., y1_+np.sin(theta1)*self._w/2.


        dx2, dy2 = df(t[::-1], args)
        n = np.hypot(dx2, dy2)
        dx2, dy2 = dx2/n, dy2/n
        x2_, y2_ = f(t[::-1], args)
        theta2 = np.angle(dx2+1j*dy2)+np.pi/2.
        x2, y2 = x2_+np.cos(theta2)*self._w/2., y2_+np.sin(theta2)*self._w/2.

        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p = np.vstack((x, y)).T

        poly = PolygonSet(polygons=[p],
                         layers=[self._layer],
                         datatypes=[self._datatype],
                         names=[self._name],
                          colors=[self._color]).translate(self.ref[0]-x[0]+self._rot(self._w/2,0,-theta1[0])[0], self.ref[1]-y[0]+self._rot(self._w/2,0,-theta1[0])[1]).rotate(self._angle - theta1[0] - np.pi/2, [self.ref[0], self.ref[1]])
        self._angle += theta1[-1] - theta1[0]
        # Calculate curve length
        def func(t, args):
            dx1, dy1 = df(t, args)
            return np.hypot(dx1, dy1)

        # Add the length of the parametric curve only if asked (default)
        if add_length:
            self.total_length += quad(func, t[0], t[-1], args=(args,))[0]

        # Add polygon only if asked (default)
        if add_polygon:
            self._add2param(self.ref[0]+x1_,
                            self.ref[1]+y1_,
                            np.linspace(0, 1, len(t))*quad(func, t[0], t[-1], args=(args,))[0])
            self.ref = (poly.polygons[0][int(len(poly.polygons[0])/2)-1] + poly.polygons[0][int(len(poly.polygons[0])/2)])/2
            self._add(poly)

            return self
        else:
            return poly



import numpy as np
import copy

from pygdsdesign import GdsLibrary, PolygonSet, offset

layer_big   = {'layer': 1, 'name' : 'Al'}
layer_small = {'layer': 2, 'name' : 'Al', 'datatype' : 1}
offset_value = 0.250

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')

# Create empty polygon to which other polygon will be added
tot = PolygonSet()

# Will contains full size design
full = PolygonSet(**layer_big)

# Will contains offset size design
o = PolygonSet(**layer_small)

# First contact line
p1 = PolygonSet(**layer_big)
p1 > (-13, 0)
p1 > (0, 10)
p1 > (4, 0)
p1 > (0, 20)
p1 > (2, 0)
p1 > (0, 30)
p1 > (1, 0)
p1 > (0, -30)
p1 > (2, 0)
p1 > (0, -20)
p1 > (4, 0)
p1 > (0, -10)

dx, dy = p1.get_size()
p1.translate(dx/2, -dy-1)

full += p1
full += copy.copy(p1).rotate(np.deg2rad(90))
full += copy.copy(p1).rotate(np.deg2rad(180))
full += copy.copy(p1).rotate(np.deg2rad(270))

o = offset(full, -offset_value, join_first=True, **layer_small)


tot += full
tot += o

# Add polygons to cell
cell.add(tot.merge().center())

# Save GDS file
lib.export_gds('structures_offset.gds')



import copy
import os
import warnings
from typing import Any, Dict, Optional

import numpy as np

from pygdsdesign.library import GdsLibrary
from pygdsdesign.polygons import Rectangle, RectangleCentered, Text
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.operation import boolean, offset, merge, addition, subtraction, inverse_polarity


def crosses(coordinates: list,
            layer: int = 1,
            datatype: int = 0,
            width: float=5,
            h_length: float=35,
            v_length: float=35) -> PolygonSet:
    """
    Returns a polygon containing crosses at all coordinates specified in the
    coordinates list.

    Args:
        coordinates (list): List of tupples with coordinats.
        layer (int, optional): gds layer of the crosses. Defaults to 1.
        datatype (int, optional): gds datatype of the crosses. Defaults to 1.
        width (float, optional): Width of the arms of a single cross.
        Defaults to 5.
        h_length (float, optional): Total horizantal length of single cross,
        includes the width of the arm. Defaults to 35.
        v_length (float, optional): Total vertical length of single cross,
        includes the width of the arm. Defaults to 35.

    Returns:
        PolygonSet: Set of polygons containing crosses at positions
        specified by coordinates.
    """
    crosses = PolygonSet(layers=[layer], datatypes=[datatype])
    for coord in coordinates:
        cr = cross(layer=layer, datatype=datatype, width=width, h_length=h_length, v_length=v_length)
        cr.translate(coord[0], coord[1])
        crosses+=cr
    return crosses


def cross(layer: int=1,
          datatype: int=1,
          width: float=5,
          h_length: float=35,
          v_length: float=35) -> PolygonSet:
    """
    Returns a cross, specified by width, h_length and v_length.

    Args:
        layer (int, optional): gds layer of the cross. Defaults to 1.
        datatype (int, optional): gds datatype of the cross. Defaults to 1.
        width (float, optional): Width of the arms of the cross. Defaults to 5.
        h_length (float, optional): Total horizontal length of the cross,
        includes the width of the arm. Defaults to 35.
        v_length (float, optional): Total vertical length of the cross,
        includes the width of the arm. Defaults to 35.

    Returns:
        Polygon: Polygon containing the cross.
    """
    cross = PolygonSet(layers=[layer], datatypes=[datatype])
    cross += Rectangle((-h_length/2, -width/2), (h_length/2, width/2),
                       layer=layer, datatype=datatype)
    cross += Rectangle((-width/2, -v_length/2), (width/2, v_length/2),
                       layer=layer, datatype=datatype)
    merge(cross.center())
    return cross


def global_marks_ebeam(w: float=10,
                       l: float=200,
                       directional_structures: bool=True,
                       directional_offset: float=40,
                       directional_structures_length: float=5,
                       directional_scaling: float=4,
                       squared_center: bool=False,
                       layer: int=1,
                       datatype: int=1,
                       color: str='',
                       name: str='') -> PolygonSet:
    """
    Function that returns an ebeam alignment mark for global alignment.
    When directional_structures is True, triangular structures pointing to the
    center of the mark are added to help finding the center of the cross with
    the ebeam system.

    Args:
        w: cross width in um.
            Defaults to 10.
        l: cross length in um.
            Defaults to 200.
        directional_structures: when True, structures are added to show the center
            of the cross, helping to find the cross on the ebeam system.
            Defaults to True.
        directional_offset: minimal distance between the directional mark and the
            cross. Should be large enough to avoid interferences while doing the
            mark detection
            Defaults to 40.
        directional_structures_length: Length of the directional mark in um.
            The mark beeing a triangle, this length corresponds to its base.
            Defaults to 5.
        directional_scaling: Scaling factor applied to the structure.
            Smaller size are closer to the center.
            Defaults to 4.
        squared_center: If True, add two hollow squares in diagonal at the center
            of the cross.
            It offers a easy target to pinpoint while doing manual alignement.
            Example below when squared cented is true:
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
                          %%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%             %%%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
                           %%%%%%%%%%%%%%%%%%%%%%%%%%%
            Defaults to False.
        layer: gds layer of chip marks.
            Defaults to 1.
        datatype: gds datatype of chip marks.
            Defaults to 1.
        color: gds color of chip marks.
            Defaults to ''.
        name: gds name of chip marks.
            Defaults to ''.
    """

    # Make the crosse
    cross = PolygonSet()
    cross += Rectangle((0, 0), (w, l), layer=layer, datatype=datatype, color=color, name=name).center()
    cross += Rectangle((0, 0), (l, w), layer=layer, datatype=datatype, color=color, name=name).center()

    # By default the total structure is just the cross
    tot = copy.deepcopy(cross)

    if squared_center:
        temp = Rectangle((0, 0), (w/2, w/2))
        temp += Rectangle((0, 0), (-w/2, -w/2))
        tot = boolean(cross,
                      temp,
                      'xor', layer=layer, datatype=datatype, color=color, name=name)

        # If the boolean failed, we return the cross without error and inform the user
        if tot is None:
            tot = cross
            warnings.warn('You asked for crossed center ebeam marked but it failed. Please check your input dimensions.')


    if directional_structures:
        temp = PolygonSet()

        # Create a default triangle with the proper orientation
        t = PolygonSet([[(0, 0), (directional_structures_length, 0), (directional_structures_length/2, 2*directional_structures_length)]])
        t.rotate(np.pi/2, t.get_center())

        # Add many triangles with the proper rotation in 10 concentrics circles
        for r, s in zip(np.linspace(directional_offset, l*0.75, 10),
                        np.linspace(1, directional_scaling, 10)):
            p = 2*np.pi*r
            nb_p = int(p/30)
            for theta in np.linspace(0, 2*np.pi, nb_p):
                z = r*np.exp(1j*theta)
                temp += copy.copy(t).scale(s).rotate(theta).translate(z.real, z.imag)

        # We remove the triangles being too close to the cross
        temp2 = boolean(temp,
                        offset(tot, directional_offset),
                        'not')

        # We remove the triangles being outside of the bounding_box of the cross
        temp3 = boolean(temp2,
                        Rectangle((-l/2, -l/2), (l/2, l/2)),
                        'and',
                        layer=layer, datatype=datatype, color=color, name=name)

        # In case the boolean operation return nothing (too small cross for instance)
        if temp3 is None:
            return PolygonSet()

        # We remove the triangles which have been cut from previous boolean
        # operation and are now too small
        temp4 = PolygonSet()
        for p in temp3.polygons:
            t=PolygonSet([p], layers=[layer], datatypes=[datatype], colors=[color], names=[name])
            if t.get_area()>0.9*directional_structures_length*directional_structures_length:
                temp4 += t
        tot += temp4

    return tot


def mark_uv(layer_uv_mask:int = 0,
            datatype_uv_mask:int = 0,
            layer_metallisation: int = 1,
            datatype_metallisation:int = 0,
            square_size: int = 30,
            size_difference: int = -3,
            window_height:int = 170,
            window_width:int = 120 ,
            nb_repetitions: int = 3,
            ) -> PolygonSet:
    """
        Returns an UV-mark.

    Args:
        layer_uv_mask (int, optional): Layer for the UV mask. Defaults to 0.
        datatype_uv_mask (int, optional): Datatype for the UV mask. Defaults to 0.
        layer_metallisation (int, optional): Layer for the mark on the chip. Defaults to 1.
        datatype_metallisation (int, optional): Datatype for the mark on the chip. Defaults to 0.
        square_size (int, optional): Size of the square marks on the chip, in um. Defaults to 30.
        size_difference (int, optional): Size difference of the squares between the marks on the wafer and the UV mask, in um. Defaults to -3.
        window_height (int, optional): Height of the surrounding window on the UV mask, in um. Defaults to 170.
        window_width (int, optional): Width of the surrounding window on the UV mask, in um. Defaults to 120.
        nb_repetitions (int, optional): Number of pairs of square marks. Defaults to 3.

    Returns:
        PolygonSet: Set of polygons containing UV marks.
    """
    square = Rectangle(point1=[0,0], point2=[square_size, square_size], layer=layer_metallisation, datatype=datatype_metallisation)
    unit = square + copy.deepcopy(square).translate(0, 50)
    unit.center()
    dummy_unit = offset(unit, distance=size_difference)
    window = Rectangle(point1=[0,0], point2=[window_width, window_height], layer=layer_uv_mask, datatype=datatype_uv_mask).center()
    frame1 = offset(window, distance=-5*size_difference)
    frame2 = offset(window, distance=-4*size_difference)
    frame = boolean(operand1=frame1, operand2=frame2, operation='not', layer=layer_uv_mask, datatype=datatype_uv_mask)
    window = boolean(operand1=window, operand2=dummy_unit, operation='not', layer=layer_uv_mask, datatype=datatype_uv_mask)
    unit += window
    unit += frame
    units = PolygonSet()
    for i in range(nb_repetitions):
        units += copy.deepcopy(unit).translate(i*120, 0)

    return merge(units.center())


def chip_marks_ebeam(layer: int=1,
                     datatype: int=1) -> PolygonSet:
    """
    Returns a set of ebeam chip marks.

    Args:
        layer (int, optional): gds layer of chip marks. Defaults to 1.
        datatype (int, optional): gds datatype of chip marks. Defaults to 1.

    Returns:
        PolygonSet: Set of polygons containing ebeam chip marks.
    """
    cross = Rectangle((-7.5, -0.5), (7.5, 0.5),
                            layer=layer, datatype=datatype)
    cross += Rectangle((-0.5, -7.5), (0.5, 7.5),
                             layer=layer, datatype=datatype)
    crosses = PolygonSet()
    crosses += copy.copy(cross).translate(-20, -20)
    crosses += copy.copy(cross).translate(-20, 20)
    crosses += copy.copy(cross).translate(20, 20)
    crosses += copy.copy(cross).translate(20, -20)
    all_crosses = copy.copy(crosses)
    all_crosses += copy.copy(crosses).translate(132.5, 127.5)
    all_crosses += copy.copy(crosses).translate(-132.5, 127.5)
    all_crosses += copy.copy(crosses).translate(132.5, -127.5)
    all_crosses += copy.copy(crosses).translate(-132.5, -127.5)
    return all_crosses


def chip_marks_laser(layer: int=1,
                     datatype: int=1,
                     color: str='',
                     name: str='',
                     only_square: bool=False) -> PolygonSet:
    """
    Returns a set of alignement marks use by some people in optical lithography.
    Consists of a cross with a small square:

                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
            @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
            @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
            @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
            @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@
                              @@@@@@@@@@                  @@@@@@@@@
                              @@@@@@@@@@                  @@@@@@@@@
                              @@@@@@@@@@                  @@@@@@@@@
                              @@@@@@@@@@                  @@@@@@@@@
                                                 @@@@@@@@@
                                                 @@@@@@@@@
                                                 @@@@@@@@@
                                                 @@@@@@@@@


    Args:
        layer: gds layer of chip marks. Defaults to 1.
        datatype: gds datatype of chip marks. Defaults to 1.
        color: gds color of chip marks. Defaults to ''.
        name: gds name of chip marks. Defaults to ''.
        only_square: If True, return only the square of the alignement mark.
            Defaults to False.

    Returns:
        PolygonSet: Set of polygons containing laser chip marks.
    """

    # cross
    # The cross is return only if only_square is False (default)
    if only_square:
        p = PolygonSet([[(0, 0)]], [layer], [datatype], [name], [color])
    else:
        p = PolygonSet([[( -5, -10),
                         ( -5,   0),
                         (-15,   0),
                         (-15,   5),
                         ( -5,   5),
                         ( -5,  15),
                         (  0,  15),
                         (  0,   5),
                         (10,    5),
                         (10,    0),
                         (  0,   0),
                         (  0,-10)]], [layer], [datatype], [name], [color])

    ## 2 small squares
    # bottom left
    s1 = PolygonSet([[( 5, -15),
                  ( 5, -10),
                  (10, -10),
                  (10, -15)]], [layer], [datatype], [name], [color])

    # top right
    s2 = PolygonSet([[(10, -10),
                  (10,  -5),
                  (15,  -5),
                  (15, -10)]], [layer], [datatype], [name], [color])

    return p+s1+s2


def crossover(layer_dielectric: int=1,
              dt_dielectric_undercut: int=1,
              layer_metal: int=2,
              name_metal: str='',
              name_dielectric: str='',
              m: float=1,
              w: float=6,
              l: float=120,
              u: float=0) -> PolygonSet:
    """
    Returns a single crossover. Size of the dielectric is in all directions
    bigger by "m" um compared to metal bridge.

    Parameters
    ----------
    layer_dielectric : int, optional
        gds layer of the dielectirc, by default 1
    layer_metal : int, optional
        gds layer of the metal bridge, by default 2
    name_metal : str, optional
        gds layer name of the metal bridge, by default ''
    name_dielectric : str, optional
        gds layer name of the dielectirc, by default ''
    dt_dielectric_undercut : int, optional
        gds datatype of the undercut for the dielectric, by default 1
    m : float, optional
        margin between dielectric and metal in units of um, by default 1
    w : float, optional
        width of the metal crossover in units of um, by default 6
    l : float, optional
        length of the metal crossover in units of um, by default 120
    u : float, optional
        undercut of dielectric in units of um, by default 1

    Returns:
        PolygonSet: Set of polygons containing the crossover.
    """
    # define the metal part
    met = PolygonSet()
    met += Rectangle((0, 0), (4*m + w, 8*m + 3*w), layer=layer_metal, name=name_metal)
    met += Rectangle((4*m + w, 4*m + w), (4*m + w + l, 4*m + 2*w), layer=layer_metal, name=name_metal)
    met += Rectangle((4*m + w + l, 0), (8*m + 2*w + l, 8*m + 3*w), layer=layer_metal, name=name_metal)
    # define the dielectric part
    dielec = PolygonSet()
    dielec += Rectangle((m, m), (3*m + w, 7*m + 3*w), layer=layer_dielectric, name=name_dielectric)
    dielec += Rectangle((3*m + w, 3*m + w), (5*m + w + l, 5*m + 2*w), layer=layer_dielectric, name=name_dielectric)
    dielec += Rectangle((5*m + w + l, m), (7*m + 2*w + l, 7*m + 3*w), layer=layer_dielectric, name=name_dielectric)
    # define the mask to generate the undercut
    undercut = PolygonSet()
    undercut += Rectangle((m - u, m - u), (3*m + w + u, 7*m + 3*w + u), layer=layer_dielectric, name=name_dielectric, datatype=dt_dielectric_undercut)
    undercut += Rectangle((3*m + w - u, 3*m + w - u), (5*m + w + l + u, 5*m + 2*w + u), layer=layer_dielectric, name=name_dielectric, datatype=dt_dielectric_undercut)
    undercut += Rectangle((5*m + w + l - u, m - u), (7*m + 2*w + l + u, 7*m + 3*w + u), layer=layer_dielectric, name=name_dielectric, datatype=dt_dielectric_undercut)
    # get undercut
    temp = boolean(undercut, dielec,
                   'not',
                   layer=layer_dielectric,
                   datatype=dt_dielectric_undercut,
                   name=name_dielectric)

    tot = PolygonSet()
    tot+=met
    tot+=dielec
    if temp is not None: # in case undercut is u=0
        tot+=temp
    return merge(tot).center()


def daisychain(num: int=5,
               layer_dielectric: int=1,
               dt_dielectric_undercut: int=1,
               layer_metal: int=2,
               layer_NbN_etch: int=3,
               name_metal: str='',
               name_dielectric: str='',
               name_NbN_etch: str='',
               m: float=1,
               w: float=4,
               l: float=30,
               d: float=3,
               gap: float=50,
               b: float=200,
               u: float=0) -> PolygonSet:
    """
    Returns a daisychain of "num" crossover in a linear chain with bond pads at each end.

    Parameters
    ----------
    num : int, optional
        Number of crossovers in a linear chain, by default 5
    layer_NbN_etch : int, optional
        gds layer of NbN etching, by default layer_NbN_etch
    layer_metal : int, optional
        gds layer of the metal bridge , by default layer_metal
    layer_dielectric : int, optional
        gds layer of the dielectric, by default layer_dielectric
    name_metal : str, optional
        gds layer name of the metal bridge, by default ''
    name_dielectric : str, optional
        gds layer name of the dielectirc, by default ''
    name_NbN_etch : str, optional
        gds layer name of the  NbN etching, by default ''
    dt_dielectric_undercut : int, optional
        gds datatype of the undercut for the dielectric, by default 1
    m : int, optional
        margin betwewn dielectric and metal in untis of um, by default 1
    w : int, optional
        width of the metal crossover in units of um, by default 4
    l : int, optional
        length of the metal crossover in units of um, by default 30
    d : float, optional
        distance between two crossovers in untis of um, by default 3
    gap : float, optional
        gap etched into NbN in untis of um, by default 50
    b : float, optional
        size of square defining the bond pad in untis of um, by default 200
    u : float, optional
        undercut of dielectric in units of um, by default 1

    Returns:
        PolygonSet: Set of polygons containing the daisychain.
    """
    # create chain of crossovers
    chain = PolygonSet()
    for i in range(num):
        unit = crossover(m=m, l=l, w=w, u=u, layer_dielectric=layer_dielectric, dt_dielectric_undercut=dt_dielectric_undercut, layer_metal=layer_metal, name_dielectric=name_dielectric, name_metal=name_metal)
        unit += Rectangle((5*m + w, 0), (3*m + w + l, 10*m + 3 * w), layer=layer_NbN_etch, name=name_NbN_etch).center()
        dx, dy = unit.get_size()
        chain += unit.translate(i * (dx + d), 0)
    ll, tr = chain.get_bounding_box()
    chain += Rectangle((ll), (tr[0], ll[1] - gap), layer=layer_NbN_etch, name=name_NbN_etch)
    chain += Rectangle((ll[0], tr[1] + gap), (tr), layer=layer_NbN_etch, name=name_NbN_etch)
    chain.center()

    # get bounding box again to add bonding pads
    ll, tr = chain.get_bounding_box()
    dx, dy = chain.get_size()
    # construct first pad
    pad = Rectangle((-gap,-gap - b/2), (0, gap + b/2), layer=layer_NbN_etch, name=name_NbN_etch)
    pad += Rectangle((-gap,-gap - b/2), (b + gap, - b/2), layer=layer_NbN_etch, name=name_NbN_etch)
    pad += Rectangle((-gap, gap + b/2), (b + gap, + b/2), layer=layer_NbN_etch, name=name_NbN_etch)
    pad += Rectangle((b + gap, gap + b/2), (b, dy/2 - gap), layer=layer_NbN_etch, name=name_NbN_etch)
    pad += Rectangle((b + gap, -gap - b/2), (b, -dy/2 + gap), layer=layer_NbN_etch, name=name_NbN_etch)

    # construct second pad and move both in correct position
    pad2 = copy.copy(pad).rotate(np.pi)
    pad2.translate(+dx/2 + b, 0)
    pad.translate(-dx/2 - b, 0)
    chain += pad
    chain += pad2
    t = Text('l = {} um, w = {} um, {}x'.format(l, w, num), 30, layer=layer_metal, name=name_metal)
    t.center().translate(0, 200)
    chain += t

    return merge(chain)


def lateqs_logo(layer: int=0,
                datatype: int=0,
                name: str='',
                color: str='',
                width: float=1000,
                center: bool=False) -> PolygonSet:
    """
    Return the LaTEQS logo in a PolygonSet.

    Args:
        layer (int, optional): layer of the return polygons. Defaults to 0.
        datatype (int, optional): datatype of the return polygons. Defaults to 0.
        width (float, optional): total width of the return polygons. Defaults to 1000.
        center (bool, optional): If the logo is centered to (0, 0).
            Defaults to False, meaning bottom-left is (0,0).

    Returns:
        (PolygonSet): LaTEQS logo in a PolygonSet.
    """

    lib = GdsLibrary()

    path = os.path.join(os.path.dirname(__file__), 'gds', 'lateqs_logo.gds')
    lib.read_gds(infile=path)

    main_cell = lib.top_level()[0]
    tot = PolygonSet()
    for p in main_cell.get_polygonsets():

        tot += PolygonSet(p.polygons,
                          layers=[layer],
                          datatypes=[datatype],
                          names=[name],
                          colors=[color])

    # Rescale logo to given width (height calculated automatically)
    w, h = tot.get_size()
    tot.scale(width/w)

    if center:
        tot.center()

    return tot


def qubit_layer_42(layer: int=42,
                   datatype: int=0) -> PolygonSet:
    """
    Generates layer 42 of the Qubit mask (post-pro alignment marks).
    The global marks are not exactley the same as on the original mask, but
    their center is the same.

    Parameters
    ----------
    layer : int, optional
        GDS layer, by default 42
    datatype : int, optional
        GDS datatype, by default 0

    Returns
    -------
    PolygonSet
        Set of polygons containing layer 42 of the qubit mask.
    """

    layer42 = PolygonSet(layers=[layer],
                         datatypes=[datatype],
                         names=['QUBIT mask marks 42'])

    # list of coordinates of global marks
    LETI_global_mark_coordinates = [(-5655, 6990), (65, 6990), (5655, 6990),
                                    (-5655, 4990), (65, 4990), (5655, 4990),
                                    (-5655, 950), (5655, 950),
                                    (-5655, -1050), (5655, -1050),
                                    (-5655, -5090), (65, -5090), (5655, -5090),
                                    (-5655, -7090), (65, -7090), (5655, -7090),]

    # list of coordinates of the center of the first chip marks in scribe
    LETI_chip_mark_coordinates = [(-1363.5, 7306.5), (1496.5, 7306.5),
                                  (-2663.5, 4286.5), (66.5, 4286.5), (2666.5, 4286.5),
                                  (-1363.5, 1266.5), (1496.5, 1266.5),
                                  (-2663.5, -1753.5), (66.5, -1753.5), (2666.5, -1753.5),
                                  (-1363.5, -4773.5), (1496.5, -4773.5)]

    # generate the global marks
    LETI_global_marks = crosses(coordinates=LETI_global_mark_coordinates,
                                layer=layer,
                                datatype=datatype,
                                width=10,
                                h_length=380,
                                v_length=1000)
    layer42 += LETI_global_marks

    # generate the chip marks (crosses and squares)
    for coord in LETI_chip_mark_coordinates:
        # generate the crosses
        LETI_chip_marks=PolygonSet(layers=[layer], datatypes=[datatype])
        LETI_chip_mark = cross(layer=layer,
                               datatype=datatype,
                               width=1,
                               h_length=16,
                               v_length=16).center()
        LETI_chip_mark += copy.copy(LETI_chip_mark).translate(40,0)
        LETI_chip_mark.center()
        LETI_chip_mark += copy.copy(LETI_chip_mark).translate(0,40)
        LETI_chip_mark.center()
        for i in range(11):
            LETI_chip_marks += copy.copy(LETI_chip_mark).translate(0, -i * 260)
        layer42 += LETI_chip_marks.translate(coord[0], coord[1])

        # generate the squares
        square = Rectangle((0,0), (8,8), layer=layer, datatype=datatype)
        squares=PolygonSet(layers=[layer], datatypes=[datatype])
        for i in range(4):
            for j in range(4):
                squares += copy.copy(square).translate(i * 16, j * 16)
        squares.center()
        squares.translate(0, -130).translate(-1.5, -1.5)
        squares_final = PolygonSet(layers=[layer], datatypes=[datatype])
        for i in range(11):
            squares_final += copy.copy(squares).translate(0, -i * 260)
        layer42 += squares_final.translate(coord[0], coord[1])

    return layer42


def resistivity_4_probes(layer: int=0,
                         name: str='',
                         datatype: int=0,
                         color: str='',
                         pad_width: float=400,
                         current_length: float=4000,
                         current_width: float=80,
                         voltage_length: float=400,
                         voltage_width: float=40,
                         gap:Optional[float]=None,
                         centered: bool=True) -> PolygonSet:
    """
    Return a structure dedicated to a 4 probes DC measurement

    Args:
        layer: Number of the metal layer.
            Defaults to 0.
        name: Name of the metal layer.
            Defaults to ''.
        color: Color of the metal layer.
            Defaults to ''.
        datatype: Datatype of the metal layer.
            Defaults to 0.
        pad_width: Width of the bonding pad. The bonding bad is a square.
            In um.
            Defaults to 400.
        current_length: Effective length of the current line.
            In um.
            This is the length measured by the voltage probe, not the total line length.
            The total line length will be current_length + 2*voltage_length
            Defaults to 4000.
        current_width: Width of the current line.
            Must be much smaller than the current length.
            In um.
            Defaults to 80.
        voltage_length: Length of the voltage probes line.
            In um.
            Defaults to 400.
        voltage_width: Width of the voltage probes line.
            Must be much smaller than the current length.
            In um.
            Defaults to 40.
        gap: If not None, return the surrounding gap through offset and boolean
            operation.
            In um.
            Defaults to None.
        centered: If True, centered the structure.
            Defaults to True.
    """

    _layer: Dict[str, Any] = {'layer'    : layer,
                              'datatype' : datatype,
                              'name'     : name,
                              'color'    : color}

    tot = PolygonSet(polygons=[[(0,0)]], **_layer)

    # Make the current line
    tot += Rectangle((0, 0), (pad_width, pad_width), **_layer)
    tot += Rectangle((0, 0), (current_length+2*voltage_length, current_width), **_layer).translate(pad_width, pad_width/2-current_width/2)
    tot += Rectangle((0, 0), (pad_width, pad_width), **_layer).translate(pad_width+current_length+2*voltage_length, 0)

    # Add the 1st voltage probe
    tot += Rectangle((0, 0), (pad_width, pad_width), **_layer).translate(pad_width+voltage_length-pad_width/2, pad_width/2+current_width/2+voltage_length)
    tot += Rectangle((0, 0), (voltage_width, voltage_length), **_layer).translate(pad_width+voltage_length-voltage_width/2, pad_width/2+current_width/2)

    # Add the 2nd voltage probe
    tot += Rectangle((0, 0), (pad_width, pad_width), **_layer).translate(pad_width+voltage_length+current_length-pad_width/2, pad_width/2+current_width/2+voltage_length)
    tot += Rectangle((0, 0), (voltage_width, voltage_length), **_layer).translate(pad_width+voltage_length+current_length-voltage_width/2, pad_width/2+current_width/2)

    # Get the surrounding gap through offset and boolean operation
    if gap is not None:
        temp = boolean(tot,
                       offset(tot,
                              gap,
                              join_first=True,
                              **_layer),
                       'xor',
                       **_layer)

        if temp is not None:
            tot = PolygonSet(polygons=temp.polygons,
                             layers=temp.layers,
                             datatypes=temp.datatypes,
                             names=temp.names,
                             colors=temp.colors)

    if centered:
        return tot.center()
    else:
        return tot


def qubit_layer_19(
    layer: int = 19,
    layer_annotation: int = 49,
    datatype: int = 0,
    datatype_annotation:int = 0
) -> PolygonSet:
    """
    Generates part of the layer 19 (via active to Metal 1) of the Qubit mask LETI.
    The device vias generated are:
    - 4G11_1
    - 4G11_2
    - 4G11_3
    - 4G21_1
    - 4G21_2
    - 4G22_1
    - 4G22_2
    - 4G23_1
    - 4G23_2
    - 8G11
    - 7G11_1
    - 7G11_2

    Each device is associated with a Text (5 micron above the top left via) with the device name in layer_annoation (layer 49 by default)

    Parameters
    ----------
    layer : int, optional
        GDS layer, by default 19
    layer_annotation: int, optinal
        GDS layer, by default 49
    datatype : int, optional
        GDS datatype, by default 0 for both layer and layer_annotations

    Returns
    -------
    PolygonSet
        Set of polygons containing part of the layer 19 of the LETI qubit mask
        + text annotation to locate devices in layer_annotation (default 49)
    """

    # dictonary with the via coordinate (center) of different devices

    devices = {
        "4G11_1": [
            (-1835.305, 4023.675),
            (-1834.96, 4023.695),
            (-1834.615, 4023.675),
            (-1835.305, 4022.745),
            (-1834.96, 4022.725),
            (-1834.615, 4022.745),
            ],
        "4G11_2": [
            (-1835.295, 3243.605),
            (-1834.96, 3243.625),
            (-1834.625, 3243.605),
            (-1835.295, 3242.745),
            (-1834.96, 3242.725),
            (-1834.625, 3242.745),
            ],
        "4G11_3": [
            (-1835.285, 2463.535),
            (-1834.96, 2463.555),
            (-1834.635, 2463.535),
            (-1835.285, 2462.745),
            (-1834.96, 2462.725),
            (-1834.635, 2462.745),
            ],
        "5G12_1":[
            (-1575.28500, 3962.61500),
            (-1574.96000, 3962.63500),
            (-1574.63500, 3962.61500),
            (-1575.28500, 3961.74500),
            (-1574.96000, 3961.72500),
            (-1574.63500, 3961.74500),
            (-1574.30500, 3961.74500),
            ],
        "5G12_2":[
            (-1575.28500, 3052.61500),
            (-1574.96000, 3052.63500),
            (-1574.63500, 3052.61500),
            (-1575.28500, 3051.74500),
            (-1574.96000, 3051.72500),
            (-1574.63500, 3051.74500),
            (-1574.30500, 3051.74500)
            ],
        "6G11_2":[
            (-1445.61500, 2854.69500),
            (-1445.28500, 2854.69500),
            (-1444.96000, 2854.71500),
            (-1444.63500, 2854.69500),
            (-1445.28500, 2853.74500),
            (-1444.96000, 2853.72500),
            (-1444.63500, 2853.74500),
            (-1444.30500, 2853.74500)
            ],
        "7G11_1": [
            (-1315.625, 3830.905),
            (-1315.295, 3830.905),
            (-1314.96, 3830.925),
            (-1314.625, 3830.905),
            (-1315.625, 3829.745),
            (-1315.295, 3829.745),
            (-1314.96, 3829.725),
            (-1314.625, 3829.745),
            (-1314.295, 3829.745),
            ],
        "7G11_2": [
            (-1315.615, 2660.775),
            (-1315.285, 2660.775),
            (-1314.96, 2660.795),
            (-1314.635, 2660.775),
            (-1315.615, 2659.745),
            (-1315.285, 2659.745),
            (-1314.96, 2659.725),
            (-1314.635, 2659.745),
            (-1314.305, 2659.745),
            ],
        "8G11": [
            (-1185.615, 3764.855),
            (-1185.615, 3764.855),
            (-1185.285, 3764.855),
            (-1184.96, 3764.875),
            (-1184.635, 3764.855),
            (-1184.305, 3764.855),
            (-1185.615, 3763.745),
            (-1185.285, 3763.745),
            (-1184.96, 3763.725),
            (-1184.635, 3763.745),
            (-1184.305, 3763.745),
            ],
        "4G21_1": [
            (2194.695, 6903.675),
            (2195.05, 6903.695),
            (2195.405, 6903.675),
            (2194.695, 6902.745),
            (2195.05, 6902.725),
            (2195.405, 6902.745),
            ],
        "4G21_2": [
            (2194.695, 5863.675),
            (2195.04, 5863.695),
            (2195.385, 5863.675),
            (2194.695, 5862.745),
            (2195.04, 5862.725),
            (2195.385, 5862.745),
            ],
        "4G22_1": [
            (2324.705, 6903.605),
            (2325.05, 6903.625),
            (2325.395, 6903.605),
            (2324.705, 6902.745),
            (2325.05, 6902.725),
            (2325.395, 6902.745),
            ],
        "4G22_2": [
            (2324.705, 5863.605),
            (2325.04, 5863.625),
            (2325.375, 5863.605),
            (2324.705, 5862.745),
            (2325.04, 5862.725),
            (2325.375, 5862.745),
            ],
        "4G23_1": [
            (2454.715, 6903.535),
            (2455.05, 6903.555),
            (2455.385, 6903.535),
            (2454.715, 6902.745),
            (2455.05, 6902.725),
            (2455.385, 6902.745),
            ],
        "4G23_2": [
            (2454.715, 5863.535),
            (2455.04, 5863.555),
            (2455.365, 5863.535),
            (2454.715, 5862.745),
            (2455.04, 5862.725),
            (2455.365, 5862.745),
            ],
        "5G23_1":[
            (2844.61500, 6850.61500),
            (2844.95000, 6850.63500),
            (2845.28500, 6850.61500),
            (2844.28500, 6849.74500),
            (2844.61500, 6849.74500),
            (2844.95000, 6849.72500),
            (2845.28500, 6849.74500)
            ],
        "5G23_2":[
            (2844.63500, 5680.61500),
            (2844.96000, 5680.63500),
            (2845.28500, 5680.61500),
            (2844.30500, 5679.74500),
            (2844.63500, 5679.74500),
            (2844.96000, 5679.72500),
            (2845.28500, 5679.74500)
            ],
        "6G22_1":[
            (3104.36500, 6784.69500),
            (3104.70000, 6784.71500),
            (3105.03500, 6784.69500),
            (3105.36500, 6784.69500),
            (3104.03500, 6783.74500),
            (3104.36500, 6783.74500),
            (3104.70000, 6783.72500),
            (3105.03500, 6783.74500)
            ],
        "6G22_2":[
            (3104.38500, 5484.69500),
            (3104.71000, 5484.71500),
            (3105.03500, 5484.69500),
            (3105.36500, 5484.69500),
            (3104.05500, 5483.74500),
            (3104.38500, 5483.74500),
            (3104.71000, 5483.72500),
            (3105.03500, 5483.74500)
            ],
        "7G22_1":[
            (3366.38500, 6719.25500),
            (3366.71500, 6719.25500),
            (3367.04000, 6719.27500),
            (3367.36500, 6719.25500),
            (3367.69500, 6719.25500),
            (3366.38500, 6718.22500),
            (3366.71500, 6718.22500),
            (3367.04000, 6718.20500),
            (3367.36500, 6718.22500)
            ],
        "8G22_1":[
            (3621.38500, 6723.85500),
            (3621.71500, 6723.85500),
            (3622.04000, 6723.87500),
            (3622.36500, 6723.85500),
            (3622.69500, 6723.85500),
            (3621.38500, 6722.74500),
            (3621.71500, 6722.74500),
            (3622.04000, 6722.72500),
            (3622.36500, 6722.74500),
            (3622.69500, 6722.74500)
            ],
    }

    annotations = PolygonSet(names=['QUBIT device names'])
    vias = PolygonSet(names=['QUBIT contacts 19'])

    for device in devices.keys():
        coordinates = devices[device]
        annotations += Text(
            device,
            20,
            position=(coordinates[0][0], coordinates[0][1] + 0.5),
            layer=layer_annotation,
            datatype=datatype_annotation,
            name='QUBIT device names'
        )
        for via_coordinates in coordinates:
            vias += RectangleCentered(
                (via_coordinates[0], via_coordinates[1]),
                0.09,
                0.09,
                layer=layer,
                datatype=datatype,
                name = 'QUBIT contacts 19'
            )

    total = vias + annotations
    return total


def dicing_saw_mark(substrate: str='si',
                    layer: int=1,
                    name: str='',
                    color: str='',
                    datatype: int=1,
                    ratio: float=5) -> PolygonSet:
    """
    Return dicing saw marks in a shape of a cross.
    Theses mark are done in such way that the blade thickness used at the BCAI
    will completely delete the mark from the leftover chips, a.k.a. the blade
    thickness corresponds to the mark width.
    Hence, for each substrate type corresponds a mark thickness:
        Si    -> 60um width
        Al2O3 -> 250um width

    Args:
        substrate: Nature of the substrate, must be: ('si', 'silicium', 'al2o3',
            'sapphire').
            Defaults to 'si'.
        layer: Number of the metal layer.
            Defaults to 0.
        name: Name of the metal layer.
            Defaults to ''.
        color: Color of the metal layer.
            Defaults to ''.
        datatype: Datatype of the metal layer.
            Defaults to 0.
        ratio: ratio length/width of the dicing saw mark.
            Defaults to 5.
    """

    if substrate.lower() in ('si', 'silicium'):
        w = 60
        l = ratio*w
    elif substrate.lower() in ('al2o3', 'sapphire'):
        w = 250
        l = ratio*w
    else:
        raise ValueError('substrate must be "si", or "al2o3"')

    s = l/2 - w/2

    t = PolygonSet(polygons   = [[(-l/2, w/2)]],
                   layers     = [layer],
                   datatypes  = [datatype],
                   colors     = [color],
                   names      = [name])
    t > ( s,  0)
    t > ( 0,  s)
    t > ( w,  0)
    t > ( 0, -s)
    t > ( s,  0)
    t > ( 0, -w)
    t > (-s,  0)
    t > ( 0, -s)
    t > (-w,  0)
    t > ( 0,  s)
    t > (-s,  0)
    t > ( 0,  w/2)

    return t


def dicing_saw_mark_hollow(substrate: str='si',
                           layer: int=1,
                           name: str='',
                           color: str='',
                           datatype: int=1,
                           ratio: float=5) -> PolygonSet:
    """
    Hollow version of `dicing_saw_mark`.
    shorter by a factor 8 to write.
    Return dicing saw marks in a shape of a cross.
    Theses mark are done in such way that the blade thickness used at the BCAI
    will completely delete the mark from the leftover chips, a.k.a. the blade
    thickness corresponds to the mark width.
    Hence, for each substrate type corresponds a mark thickness:
        Si    -> 60um width
        Al2O3 -> 250um width

    Args:
        substrate: Nature of the substrate, must be: ('si', 'silicium', 'al2o3',
            'sapphire').
            Defaults to 'si'.
        layer: Number of the metal layer.
            Defaults to 0.
        name: Name of the metal layer.
            Defaults to ''.
        color: Color of the metal layer.
            Defaults to ''.
        datatype: Datatype of the metal layer.
            Defaults to 0.
        ratio: ratio length/width of the dicing saw mark.
            Defaults to 5.
    """

    if substrate.lower() in ('si', 'silicium'):
        w = 60
        l = ratio*w
    elif substrate.lower() in ('al2o3', 'sapphire'):
        w = 250
        l = ratio*w
    else:
        raise ValueError('substrate must be "si", or "al2o3"')

    s = l/2 - w/2
    cusp = w

    t = PolygonSet(polygons   = [[(-l/2, w/2)]],
                   layers     = [layer],
                   datatypes  = [datatype],
                   colors     = [color],
                   names      = [name])
    t > ( s,  0)
    t > ( 0,  s)
    t > ( w/2,  -cusp)
    t > ( w/2,   cusp)
    t > ( 0, -s)
    t > ( s,  0)
    t > ( -cusp, -w/2)
    t > ( cusp, -w/2)
    t > (-s,  0)
    t > ( 0, -s)
    t > (-w/2,  cusp)
    t > (-w/2,  -cusp)
    t > ( 0,  s)
    t > (-s,  0)
    t > ( cusp,  w/2)

    from .operation import grid_cover

    return subtraction(t, grid_cover(t,
                                     centered=True,
                                     square_width=w/10,
                                     square_gap=2,
                                     safety_margin=2,
                                     only_square=False,
                                     hexagonal_grid=True,
                                     ),
                       layer=layer,
                       datatype=datatype,
                       name=name,
                       color=color,
                       )


def spiral(
    inner_diameter: float,
    width: float,
    spacing: float,
    nb_turn: int,
    nb_points: int=500,
    layer: int=0,
    name: str="",
    color: str="",
    datatype: int=0,
) -> PolygonSet:
    """
        Make a archimedean spiral as below

                              ******************
                          ****                ******
                        ****                      ****
                      **                            ****
                    ****                              ****
                  ****                                  **
                  **              **********            ****
                ****          ****        ****            **
                **            **            ****          **
                **          ****              **          **
                **          **              ****          **
                **          **          ******            **
                **          **                          ****          **
                **          ****                        **            **
                **            **                      ****          **
                  **          ****                  ****            **
                  **            ******            ****            ****
                  ****              **************                **
                    **                                          **
                      **                                      ****
                      ****                                  ****
                          ****                          ******
                            ******                  ******
                                ********      ********
                                      **********

        Args:
            inner_diameter: Inner diameter from which the spiral will start
            width: width of the spiral arm
            spacing: spacing between the spiral arm
            nb_turn: nb turn of the spiral
            nb_points: nb_points of the polygon making the spiral.
                Defaults to 500.
            layer: Number of the metal layer.
                Defaults to 0.
            name: Name of the metal layer.
                Defaults to ''.
            color: Color of the metal layer.
                Defaults to ''.
            datatype: Datatype of the metal layer.
                Defaults to 0.

        Returns:
            PolygonSet: A PolygonSet of the spiral.
    """


    # Parametric curve
    t = np.linspace(0, 1, nb_points)
    r = nb_turn * (spacing+width) * t + inner_diameter / 2
    theta = nb_turn * 2 * np.pi * t
    x = r * np.cos(theta)
    y = r * np.sin(theta)

    # outer curve
    x1 = x + np.cos(theta) * width / 2
    y1 = y + np.sin(theta) * width / 2

    # inner curve
    x2 = x - np.cos(theta) * width / 2
    y2 = y - np.sin(theta) * width / 2

    # combine both
    x = np.concatenate((x1, x2[::-1]))
    y = np.concatenate((y1, y2[::-1]))

    return PolygonSet(polygons=[np.vstack((x, y)).T],
                      layers=[layer],
                      names=[name],
                      colors=[color],
                      datatypes=[datatype])



def capacitance(c_arm_length:list=[18,24,36,24,18],
                c_central_width:float=2,
                arm_width:float=2,
                gap:float=0.2,
                length:float=32,
                port1:bool=False,
                port2:bool=False,
                layer:int=0,
                datatype:int=0,
                name: str='',
                color: str=''
                )-> PolygonSet:
    """
    Return a capacitance to ground.
    This "antenna-like" capacitance consists in a central conductor and "arms" perpendicular to it.
    The goal is to have a shape with a low number of squares (to reduce the inductance)
    but with a large circumference to yield a large capacitance to ground.
    Used for the butterworth filters.

    Args:
        c_arm_length: length of the arms of the capacitance. connected to the central conductor.
            Defaults to [18,24,36,24,18] [um].
        c_central_width: width of the central conductor, connecting the arms.
            Defaults to 2 um.
        arm_width: width of the arms.
            Defaults to 2 um.
        gap: distance between the the conductor and the ground plane.
            Defaults to 0.2 um.
        length: length of the central conductor, connecting the arms.
            Defaults to 32um.
        port1/2: Closing or opening the left/right termination of the central conductor, usefull to connect the capacitance to an other CPW.
            Defaults to False (close).
        layer,datatype,name,color: Used for naming the metal layer
            Defaults to 0,0,'',''.
    Returns:
        PolygonSet: Set of polygons containing the capacitance
    """

    _layer: Dict[str, Any] = {'layer'    : layer,
                              'datatype' : datatype,
                              'name'     : name,
                              'color'    : color}

    n=len(c_arm_length)
    antenna = PolygonSet()
    m=Rectangle([0,0],[length,-c_central_width])
    antenna+=m
    for i in range(n):
        m=Rectangle([0,0],[arm_width,-c_arm_length[i]])
        x=  + length/(n-1)*i
        m.translate(x,c_arm_length[i]/2 -c_central_width/2)
        antenna+=m
    r=offset(antenna,gap)

    if port1==True:
        rec=Rectangle([-gap,0],[0,-c_central_width])
        antenna = addition(antenna,rec)
    if port2==True:
        rec=Rectangle([length,0],[length+gap,-c_central_width])
        antenna = addition(antenna,rec)
    antenna= subtraction(r,antenna).translate(+gap,c_central_width/2)
    bp=r.translate(+gap,c_central_width/2)

    return antenna.change_layer(**_layer),bp



def inductance(nb_l_horizontal:int=1,
               len_l_horizontal:float=70,
               len_l_vertical:float=5,
               l_microstrip_width:float=0.5,
               layer:int=0,
               datatype:int=0,
               name: str='',
               color: str=''
               )-> PolygonSet:
    """
    Return an indutance in serie. Used for the butterworth filters.
    The inductance consists in a microstrip zig-zaging in a groundplane-free rectangle.

    Args:
        nb_l_horizontal: number of time the microstrip will go from one side to an other, the first and the last half-length microstrip dont count.
            Defaults to 1.
        len_l_horizontal: length of the microstrip going from the left (or right) to the right (or left) side.
            Defaults to 70 um.
        len_l_vertical: distance between two horizontal microstrip.
            Defaults to 5 um.
        l_microstrip_width: width of the microstrip.
            Defaults to 0.5um.
        layer,datatype,name,color: Used for naming the metal layer
            Defaults to 0,0,'',''.
    Returns:
        PolygonSet: Set of polygons containing the indutance
    """

    from pygdsdesign.transmission_lines.microstrip_polar import MicroStripPolar

    _layer: Dict[str, Any] = {'layer'    : layer,
                              'datatype' : datatype,
                              'name'     : name,
                              'color'    : color}

    dy= 5+5+ (nb_l_horizontal+2)*l_microstrip_width + len_l_vertical*(nb_l_horizontal+1)
    dx= len_l_horizontal + l_microstrip_width*20


    m=MicroStripPolar(l_microstrip_width,np.pi/2)
    m.add_line(5)
    m.add_turn(l_microstrip_width/2,np.pi/2)
    m.add_line(len_l_horizontal/2-l_microstrip_width/2)
    m.add_turn(l_microstrip_width/2,-np.pi/2)

    if nb_l_horizontal%2 == 1: #the number of horizontal repition is odd, first we use loop to create nb_l_horizontal - 1 strip, and then we add the last one.
        for i in range(int((nb_l_horizontal-1)/2)):
           m.add_line(len_l_vertical)
           m.add_turn(l_microstrip_width/2,-np.pi/2)
           m.add_line(len_l_horizontal)
           m.add_turn(l_microstrip_width/2,+np.pi/2)
           m.add_line(len_l_vertical)
           m.add_turn(l_microstrip_width/2,+np.pi/2)
           m.add_line(+len_l_horizontal)
           m.add_turn(l_microstrip_width/2,-np.pi/2)
        #last full horizotal
        m.add_line(len_l_vertical)
        m.add_turn(l_microstrip_width/2,-np.pi/2)
        m.add_line(+len_l_horizontal)
        m.add_turn(l_microstrip_width/2,+np.pi/2)
        m.add_line(len_l_vertical)
        #last half horizotal
        m.add_turn(l_microstrip_width/2,np.pi/2)
        m.add_line(len_l_horizontal/2-l_microstrip_width/2)
        m.add_turn(l_microstrip_width/2,-np.pi/2)
    else:
        for i in range(int((nb_l_horizontal)/2)):
            m.add_line(len_l_vertical)
            m.add_turn(l_microstrip_width/2,-np.pi/2)
            m.add_line(len_l_horizontal)
            m.add_turn(l_microstrip_width/2,+np.pi/2)
            m.add_line(len_l_vertical)
            m.add_turn(l_microstrip_width/2,+np.pi/2)
            m.add_line(+len_l_horizontal)
            m.add_turn(l_microstrip_width/2,-np.pi/2)
        #last half horizotal
        m.add_line(len_l_vertical)
        m.add_turn(l_microstrip_width/2,-np.pi/2)
        m.add_line(len_l_horizontal/2-l_microstrip_width/2)
        m.add_turn(l_microstrip_width/2,np.pi/2)

    m.add_line(5)
    m.translate(dx/2,0)
    rec=Rectangle([0,0],[dx,dy])
    induc=subtraction(rec,m).translate(-dx/2,0)

    return induc.change_layer(**_layer)


def butterworth_filter(central_conductor_width:float=4,
                       central_conductor_gap:float=0.2,
                       sep_bot_top:float=6,
                       sep_antenna_indutance:float=6,
                       sep_inductance_antenna:float=6,
                       sep_antenna_central:float=6,#
                       nb_l_horizontal:int=1,
                       len_l_horizontal:float=70,
                       len_l_vertical:float=5,
                       l_microstrip_width:float=0.5,#
                       c_arm_length1:list=[18,26,34,34,26,18],
                       c_arm_length2:list=[18,28,38,38,28,18],
                       c_arm_length3:list=[34,26,18],
                       c_central_width:float=2,
                       arm_width:float=2,
                       gap:float=0.2,
                       length:list=[14,32],
                       layer:float=0,
                       datatype:int=0,
                       name: str='',
                       color: str=''
                       )-> PolygonSet:
    """
    Return a 5th-order, cauer topology, butterworth filter and its bounding polygons.
    see \pygdsdesign\examples\butterworth_filter_parameters.png for a graphical reprensations of the parameters.

    Args:
        central_conductor_width: width of the central CPW.
            Defaults to 4 um
        central_conductor_gap: gap of the central CPW.
            Defaults to 0.2um
        sep_bot_top: distance added at the start and at the end of the filter. used to separate the filter from other elements.
            Defaults to 6um.
        sep_antenna_indutance: distance between the first capacitance and the first indutance, and between the second inductance and the third capacitance.
            Defaults to 6um.
        sep_inductance_antenna: distance between the first inductance and the second capacitance and between the second capacitance and the second inductance.
            Defaults to 6um.
        sep_antenna_central: distance between the capacitance and the central conductor.
            Defaults to 6um.
        nb_l_horizontal: inductance parameters. number of time the microstrip will go from one side to an other, the first and the last half-length microstrip dont count.
            Defaults to 1.
        len_l_horizontal: inductance parameters. length of the microstrip going from the left (or right) to the right (or left) side.
            Defaults to 70 um.
        len_l_vertical: inductance parameters. distance between two horizontal microstrip.
            Defaults to 5 um.
        l_microstrip_width: inductance parameters. width of the microstrip.
            Defaults to 0.5um.
        c_arm_length1/2/3: capacitance parameters. length of the arm of the first and third capacitance/ the first and the third part of the second capacitance / the second part of the second capacitance.
            Defaults to [18,24,36,24,18] [um].
        c_central_width: capacitance parameters. width of the central conductor, connecting the arms.
            Defaults to 2 um.
        arm_width: capacitance parameters. width of the arms.
            Defaults to 2 um.
        gap: capacitance parameters. distance between the the conductor and the ground plane.
            Defaults to 0.2 um.
        length: capacitance parameters. lengths of the central conductor of the capacitance, connecting the arms.
            Defaults to [14,32]um.
        layer,datatype,name,color: Used for naming the metal layer
            Defaults to 0,0,'',''.
    Returns:
        PolygonSet: Set of polygons containing the butterworth filter
        PolygonSet: Set of polygons containing the butterworth filter bounding polygons.
    """
    from pygdsdesign.transmission_lines.cpw_polar import CPWPolar

    _layer: Dict[str, Any] = {'layer'    : layer,
                              'datatype' : datatype,
                              'name'     : name,
                              'color'    : color}

    tot = PolygonSet(polygons=[[(0,0)]])
    bp = PolygonSet()

    capa1,bp1=capacitance(c_arm_length=c_arm_length3,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[0],port1=True,port2=False)
    capa2,bp2=capacitance(c_arm_length=c_arm_length3,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[0],port1=True,port2=False)
    capa1=capa1.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap,max(c_arm_length3)/2+gap + sep_bot_top)
    tot+=capa1
    tot+=capa2.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap),max(c_arm_length3)/2+gap+ sep_bot_top)
    bp+=bp1.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap,max(c_arm_length3)/2+gap + sep_bot_top)
    bp+=bp2.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap),max(c_arm_length3)/2+gap+ sep_bot_top)

    #bp+=Rectangle(capa1.get_bounding_box()[0],capa1.get_bounding_box()[1])
    #bp+=Rectangle(capa2.get_bounding_box()[0],capa2.get_bounding_box()[1])

    central=CPWPolar(central_conductor_width,central_conductor_gap,np.pi/2)
    central.add_line(max(c_arm_length3)+2*gap +sep_bot_top)
    central.add_line(sep_antenna_indutance)
    bp+=Rectangle(central.get_bounding_box()[0],central.get_bounding_box()[1])

    hori=CPWPolar(c_central_width,gap,np.pi)
    hori.add_line(central_conductor_gap*2 + central_conductor_width + 2* sep_antenna_central).translate(capa1.get_center()[0]-capa1.get_size()[0]/2, capa1.get_center()[1] )
    bp+=Rectangle(hori.get_bounding_box()[0],hori.get_bounding_box()[1])

    tot+=subtraction(central,boolean(inverse_polarity(hori,safety_marge=0),central,'and',precision=0.0001),precision=0.00001)
    tot+=subtraction(hori,boolean(inverse_polarity(central,safety_marge=0),hori,'and',precision=0.0001),precision=0.00001)


    i1=inductance(nb_l_horizontal=nb_l_horizontal,len_l_horizontal=len_l_horizontal,len_l_vertical=len_l_vertical,l_microstrip_width=l_microstrip_width)
    tot+=i1.translate(0,central.ref[1])
    bp+=Rectangle(i1.get_bounding_box()[0],i1.get_bounding_box()[1])

    central=CPWPolar(central_conductor_width,central_conductor_gap,np.pi/2)
    central.add_line(sep_inductance_antenna)
    tot+=central.translate(0,i1.get_center()[1]+i1.get_size()[1]/2)
    bp+=Rectangle(central.get_bounding_box()[0],central.get_bounding_box()[1])

    c1,b1=capacitance(c_arm_length=c_arm_length1,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[1],port1=True,port2=False)
    c2,b2=capacitance(c_arm_length=c_arm_length1,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[1],port1=True,port2=False)
    c3,b3=capacitance(c_arm_length=c_arm_length2,c_central_width=central_conductor_width,arm_width=arm_width,gap=gap,length=length[1],port1=True,port2=True)
    c4,b4=capacitance(c_arm_length=c_arm_length1,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[1],port1=True,port2=False)
    c5,b5=capacitance(c_arm_length=c_arm_length1,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[1],port1=True,port2=False)

    tot+=c1.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap, central.ref[1] +max(c_arm_length1)/2 + gap)
    tot+=c2.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap) ,central.ref[1] +max(c_arm_length1)/2 + gap )
    c3=c3.rotate(np.pi/2).translate(0, c1.get_center()[1]+c1.get_size()[1]/2 )
    tot+=c4.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap, c3.get_center()[1]+c3.get_size()[1]/2  +max(c_arm_length1)/2 + gap)
    tot+=c5.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap) ,c3.get_center()[1]+c3.get_size()[1]/2  +max(c_arm_length1)/2 + gap)

    bp+=b1.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap, central.ref[1] +max(c_arm_length1)/2 + gap)
    bp+=b2.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap) ,central.ref[1] +max(c_arm_length1)/2 + gap )
    bp+=b3.rotate(np.pi/2).translate(0, c1.get_center()[1]+c1.get_size()[1]/2 )
    bp+=b4.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap, c3.get_center()[1]+c3.get_size()[1]/2  +max(c_arm_length1)/2 + gap)
    bp+=b5.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap) ,c3.get_center()[1]+c3.get_size()[1]/2  +max(c_arm_length1)/2 + gap)

    central1=CPWPolar(central_conductor_width,central_conductor_gap,np.pi/2)
    central1.add_line(c1.get_size()[1]).translate(0,central.ref[1])

    central2=CPWPolar(central_conductor_width,central_conductor_gap,np.pi/2)
    central2.add_line(c3.get_size()[1]).translate(0,central1.ref[1])

    central3=CPWPolar(central_conductor_width,central_conductor_gap,np.pi/2)
    central3.add_line(c5.get_size()[1]+ sep_inductance_antenna).translate(0,central2.ref[1])

    hori1=CPWPolar(c_central_width,gap,np.pi)
    hori1.add_line(central_conductor_gap*2 + central_conductor_width + 2* sep_antenna_central).translate(c1.get_center()[0]-c1.get_size()[0]/2, c1.get_center()[1] )

    hori2=CPWPolar(c_central_width,gap,np.pi)
    hori2.add_line(central_conductor_gap*2 + central_conductor_width + 2* sep_antenna_central).translate(c4.get_center()[0]-c4.get_size()[0]/2, c4.get_center()[1] )

    tot+=subtraction(central1,boolean(inverse_polarity(hori1,safety_marge=0),central1,'and',precision=0.0001),precision=0.00001)
    tot+=subtraction(hori1,boolean(inverse_polarity(central1,safety_marge=0),hori1,'and',precision=0.0001),precision=0.00001)

    tot+=subtraction(central2,boolean(inverse_polarity(c3,safety_marge=0),central2,'and',precision=0.0001),precision=0.00001)
    tot+=subtraction(c3,boolean(inverse_polarity(central2,safety_marge=0),c3,'and',precision=0.0001),precision=0.00001)

    tot+=subtraction(central3,boolean(inverse_polarity(hori2,safety_marge=0),central3,'and',precision=0.0001),precision=0.00001)
    tot+=subtraction(hori2,boolean(inverse_polarity(central3,safety_marge=0),hori2,'and',precision=0.0001),precision=0.00001)

    bp+=Rectangle(central1.get_bounding_box()[0],central1.get_bounding_box()[1])
    bp+=Rectangle(central2.get_bounding_box()[0],central2.get_bounding_box()[1])
    bp+=Rectangle(central3.get_bounding_box()[0],central3.get_bounding_box()[1])
    bp+=Rectangle(hori1.get_bounding_box()[0],hori1.get_bounding_box()[1])
    bp+=Rectangle(hori2.get_bounding_box()[0],hori2.get_bounding_box()[1])


    i2=inductance(nb_l_horizontal=nb_l_horizontal,len_l_horizontal=len_l_horizontal,len_l_vertical=len_l_vertical,l_microstrip_width=l_microstrip_width).translate(0,central3.ref[1])
    tot+=i2
    bp+=Rectangle(i2.get_bounding_box()[0],i2.get_bounding_box()[1])

    capa1,bp1=capacitance(c_arm_length=c_arm_length3,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[0],port1=True,port2=False)
    capa2,bp2=capacitance(c_arm_length=c_arm_length3,c_central_width=c_central_width,arm_width=arm_width,gap=gap,length=length[0],port1=True,port2=False)
    tot+=capa1.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap,i2.get_center()[1]+i2.get_size()[1]/2+ max(c_arm_length3)/2+gap + sep_antenna_indutance)
    tot+=capa2.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap),i2.get_center()[1]+i2.get_size()[1]/2 + max(c_arm_length3)/2+gap+ sep_antenna_indutance)
    bp+=bp1.translate(sep_antenna_central + central_conductor_width/2 +central_conductor_gap,i2.get_center()[1]+i2.get_size()[1]/2+ max(c_arm_length3)/2+gap + sep_antenna_indutance)
    bp+=bp2.rotate(np.pi).translate(-(sep_antenna_central + central_conductor_width/2 +central_conductor_gap),i2.get_center()[1]+i2.get_size()[1]/2 + max(c_arm_length3)/2+gap+ sep_antenna_indutance)

    central=CPWPolar(central_conductor_width,central_conductor_gap,np.pi/2)
    central.add_line(sep_antenna_indutance+ sep_bot_top + max(c_arm_length3) + gap*2).translate(0,i2.get_center()[1]+i2.get_size()[1]/2)
    bp+=Rectangle(central.get_bounding_box()[0],central.get_bounding_box()[1])

    hori=CPWPolar(c_central_width,gap,np.pi)
    hori.add_line(central_conductor_gap*2 + central_conductor_width + 2* sep_antenna_central).translate(capa1.get_center()[0]-capa1.get_size()[0]/2, capa1.get_center()[1] )
    bp+=Rectangle(hori.get_bounding_box()[0],hori.get_bounding_box()[1])

    tot+=subtraction(central,boolean(inverse_polarity(hori,safety_marge=0),central,'and',precision=0.0001),precision=0.00001)
    tot+=subtraction(hori,boolean(inverse_polarity(central,safety_marge=0),hori,'and',precision=0.0001),precision=0.00001)

    return merge(tot.change_layer(**_layer)),bp


import numpy as np
import time

from pygdsdesign import GdsLibrary, Text, MicroStrip, inverse_polarity


start = time.time()

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('FIRST')


# Test basic class
m = MicroStrip(width=2, layer=6, name='test1')
m.add_line(0, 8)
m.add_line(8, 0)
m.add_line(0, 10)
m.add_line(-16, 0)
m.add_line(0, -6)
m.add_line(4, 0)
m.add_line(0, -16)
m.add_line(20, 0)
m.add_turn(15, 'lt')
m.add_line(0, 5)

m.add_line(5, 0)
m.add_line(20, 0)
m.add_turn(15, 'lt')
m.add_turn(5, 'lb')
m.add_line(0, -5)

m.add_line(0, -10)
m.add_turn(5, 'tl')
m.add_line(-5, 0)

m.add_line(0, -10)
m.add_turn(5, 'tr')
m.add_line(5, 0)

m.add_line(0, -10)
m.add_line(-10, 0)
m.add_turn(5, 'rb')
m.add_line(0, -5)

m.add_line(0, -10)
m.add_line(-10, 0)
m.add_turn(5, 'rt')
m.add_line(0, 5)

m.add_line(0, 10)
m.add_turn(5, 'br')
m.add_line(5, 0)

m.add_line(0, 60)
m.add_turn(5, 'bl')
m.add_line(-5, 0)
m.add_line(-5, 0)

m.add_taper(-10, 0, 6)
m.add_line(-8, 0)
m.add_line(-8, 0)

m.add_turn(4, 'rt')
m.add_turn(4, 'br')
m.add_taper(10, 0, 6)
m.add_line(8, 0)
m.add_line(8, 0)
m.add_taper(10, 0, 2)
m.add_line(8, 0)
m.add_line(8, 0)

m.add_turn(4, 'lt')
m.add_line(0, 8)
m.add_line(0, 8)
m.add_taper(0, 4, 5)
m.add_line(0, 8)
m.add_line(0, 8)

m.add_turn(4, 'br')
m.add_turn(4, 'lb')
m.add_line(0, -8)
m.add_line(0, -8)
m.add_taper(0, -4, 10)
m.add_line(0, -8)
m.add_line(0, -8)
m.add_turn(10, 'tr')
m.add_line(10, 0)
m.add_taper_cosec(10, 0, 2)
m.add_line(5, 0)
m.add_taper_cosec(10, 0, 5)
m.add_line(5, 0)

m.add_turn(5, 'lt')
m.add_line(0, 5)
m.add_taper_cosec(0, 10, 1)
m.add_taper_cosec(0, 12, 6)
m.add_line(0, 5)
m.add_taper_cosec(0, 10, 20)
m.add_taper_cosec(0, 12, 2)
m.add_line(0, 11.98)
m.add_turn(5, 'bl')
m.add_line(-5, 0)
m.add_line(-5, 0)
m.add_turn(5, 'rt')
m.add_turn(5, 'bl')
m.add_turn(5, 'rt')
m.add_turn(5, 'bl')
m.add_turn(5, 'rt')
m.add_turn(5, 'bl')
m.add_turn(5, 'rt')
m.add_turn(5, 'bl')
m.add_turn(5, 'rt')
m.add_turn(5, 'bl')
m.add_turn(5, 'rt')
m.add_turn(5, 'bl')
m.add_line(-1, 0)
m.add_taper_cosec(-10, 0, 5.36)
m.add_taper_cosec(-13.56, 0, 11.98)
m.add_taper_cosec(-8, 0, 2)
m.add_line(-1, 0)
m.add_turn(5, 'rb')
m.add_line(0, -2)
m.add_taper_cosec(0, -5, 4)
m.add_taper_cosec(0, -6, 6.89)
m.add_taper_cosec(0, -3, 2, 0.5)
m.add_line(0, -2)

m.layer=2
m.add_line(0, -3)

m = inverse_polarity(m)
m.translate(100, -40)

m.rotate(25, center=m.get_center())

cell.add(m)




# Test circular turn, should form a close loop
m = MicroStrip(2, 0, name='test2')
m.add_line(0, 10)
m.add_turn(50, 'br')
m.add_line(10, 0)

m.add_turn(50, 'lt')
m.add_line(0, 10)
m.add_turn(50, 'br')
m.add_line(10, 0)
m.add_turn(50, 'lb')
m.add_line(0, -10)
m.add_turn(50, 'tr')
m.add_line(10, 0)
m.add_turn(50, 'lb')
m.add_line(0, -10)
m.add_turn(50, 'tl')
m.add_line(-10, 0)
m.add_turn(50, 'rb')
m.add_line(0, -10)
m.add_turn(50, 'tl')
m.add_line(-10, 0)
m.add_turn(50, 'rt')
m.add_line(0, 10)
m.add_turn(50, 'bl')
m.add_line(-10, 0)
m.add_turn(50, 'rt')

m.center()
m += Text('Circular loop', 50, (0, 0), name='test2').center().translate(0, 250)

cell.add(m.translate(-500, 500))






# Test serpentine
m = MicroStrip(2, 0, datatype=6, name='test3')

m.add_line(250, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='right')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='right')
m.add_line(0, 250)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='top')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='top')
m.add_line(-500, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='left')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='left')
m.add_line(0, -500)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='bottom')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='bottom')

m.center()
m += Text('Test serpentine circular', 50, (0, 0), datatype=6, name='test3').center().translate(0, 600)

cell.add(m.translate(1500, 1500))



# Test serpentine
m = MicroStrip(2, 0, datatype=6, name='test3')

m.add_line(250, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='right', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='right', turn='fresnel')
m.add_line(0, 250)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='top', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='top', turn='fresnel')
m.add_line(-500, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='left', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='left', turn='fresnel')
m.add_line(0, -500)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='bottom', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='bottom', turn='fresnel')

m.center()
m += Text('Test serpentine fresnel', 50, (0, 0), datatype=6, name='test3').center().translate(0, 600)

cell.add(m.translate(0, 1500))




# Test spiral
def spiral(u, args):

    r = args[1]*2.*args[2]*u + args[0]/2.
    theta = args[1]*2.*np.pi*u
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y

def dspiral_dt(u, args):

    r = args[1]*2.*args[2]*u
    theta = args[1]*2.*np.pi*u
    dx_dt = -args[1]*2.*np.pi*np.sin(theta)*r + np.cos(theta)*args[1]*2.*args[2]
    dy_dt =  args[1]*2.*np.pi*np.cos(theta)*r + np.sin(theta)*args[1]*2.*args[2]
    return dx_dt, dy_dt


t = np.linspace(0, 1, 10000)
inner_diameter = 10
nb_turn = 40
spiral_spacing = 3.
spiral_width = 2.

m = MicroStrip(width=spiral_width, layer=3, name='test4')
m.add_parametric_curve(spiral, dspiral_dt, t,
                                args=(inner_diameter, nb_turn, spiral_spacing))
m.add_line(0, 100)

m.center()
m += Text('Test spiral', 50, (0, 0), layer=3, name='test4').center().translate(0, 400)

cell.add(m.translate(0, -750))





# Test Fresnel turn, should form a close loop
m = MicroStrip(2, layer=6, name='test5')
m.add_line(0, 10)
m.add_turn(50, 'br')
m.add_line(10, 0)

m.add_fresnel_turn(50, 'lt')
m.add_line(0, 10)
m.add_fresnel_turn(50, 'br')
m.add_line(10, 0)
m.add_fresnel_turn(50, 'lb')
m.add_line(0, -10)
m.add_fresnel_turn(50, 'tr')
m.add_line(10, 0)
m.add_fresnel_turn(50, 'lb')
m.add_line(0, -10)
m.add_fresnel_turn(50, 'tl')
m.add_line(-10, 0)
m.add_fresnel_turn(50, 'rb')
m.add_line(0, -10)
m.add_fresnel_turn(50, 'tl')
m.add_line(-10, 0)
m.add_fresnel_turn(50, 'rt')
m.add_line(0, 10)
m.add_fresnel_turn(50, 'bl')
m.add_line(-10, 0)
m.add_turn(50, 'rt')

m.center()
m += Text('Fresnel loop', 50, (0, 0), layer=6, name='test5').center().translate(0, 250)

cell.add(m.translate(500, 500))



lib.export_gds('microstrip.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')



import time

import numpy as np

from pygdsdesign import GdsLibrary
from pygdsdesign.shapes import (
    cross,
    crosses,
    global_marks_ebeam,
    chip_marks_ebeam,
    chip_marks_laser,
    dicing_saw_mark,
    dicing_saw_mark_hollow,
    mark_uv,
)

start = time.time()

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('FIRST')

# cross is a generic function to generate a simple cross, by default it is centered
tot = cross(layer=1,
            datatype=1,
            width=10,
            h_length=50,
            v_length=50
            )

# crosses returns a simple cross at each position specified in the coordinates list
tot += crosses(coordinates=[(-200, -200), (-200, 200), (200, 200), (200, -200)])

# global ebeam marks generates a cross with additional features if wanted such as structures that help to find the center (directional structure) and a clear center point (squared_center)
tot+=global_marks_ebeam(squared_center=True).translate(-500, -500)

# chip marks ebeam generate a set of four crosses we use generally with our e-beam system
tot+=chip_marks_ebeam().translate(-500, 500)

# chip marks laser generate a set of four crosses we use generally with our laser system (the location is the center of the two squares that touch)
tot+=chip_marks_laser().translate(500, 500)

# alignment marks for UV lithography
tot+=mark_uv(layer_uv_mask=10,
             layer_metallisation=11
             ).translate(500, -500)

# handy dicing saw marks that will be entierly removed by the dicing process (size correspondsto the blade width, which depens on the material)
tot+=dicing_saw_mark().translate(-1500, -1500)
tot+=dicing_saw_mark().translate(-1500, 1500)
tot+=dicing_saw_mark(substrate='sapphire').translate(1500, -1500)
tot+=dicing_saw_mark(substrate='sapphire').translate(1500, 1500)

tot+=dicing_saw_mark_hollow().translate(-4000, -4000)
tot+=dicing_saw_mark_hollow().translate(-4000, 4000)
tot+=dicing_saw_mark_hollow(substrate='sapphire').translate(4000, -4000)
tot+=dicing_saw_mark_hollow(substrate='sapphire').translate(4000, 4000)

cell.add(tot)

lib.export_gds('marks.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')


from pygdsdesign import *
import json
import os
import pytest

PATH = os.path.dirname(os.path.realpath(__file__))


with open(os.path.join(PATH, 'polygons.json')) as f:
    polygons = json.load(f)


with open(os.path.join(PATH, 'results.json')) as f:
    results = json.load(f)


@pytest.fixture
def square():
    return PolygonSet(polygons['square']['points'])


@pytest.fixture
def cross():
    return PolygonSet(polygons['cross']['points'])


@pytest.fixture
def triangle():
    return PolygonSet(polygons['triangle']['points'])


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_boolean(square,cross,triangle):

    # Check basic NOT operation
    result = boolean(square,cross, "not")
    assert np.isclose(np.asarray(results['boolean']['1']), np.asarray(result.polygons)).all()

    # Check basic NOT followed by OR
    temp = boolean(square,cross, "not")
    result = boolean(temp,triangle, "or")
    assert np.isclose(np.asarray(results['boolean']['2']), np.asarray(result.polygons)).all()

    # Check NOT with the same polygon
    result = boolean(cross,cross, "not")
    assert result == None

    # Check AND with the same polygon
    result = boolean(cross,cross, "and")
    assert np.isclose(np.asarray(result.polygons), np.asarray(result.polygons)).all()

    # Check OR with the same polygon
    result = boolean(cross,cross, "or")
    assert np.isclose(np.asarray(result.polygons), np.asarray(result.polygons)).all()

    # Check a basic XOR
    result = boolean(cross, triangle, "xor")
    for i in range(len(results['boolean']['3'])):
        assert np.isclose(np.asarray(results['boolean']['3'][i]), np.asarray(result.polygons[i])).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_offset(square,cross,triangle):

    # Check a basic offset
    result = offset(cross, 10)
    assert np.isclose(np.asarray(results['offset']['1']), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_inverse_polarity(square,cross,triangle):

    # Check a basic inverse_polarity
    result = inverse_polarity(cross)
    for i in range(len(results['inverse_polarity']['1'])):
        assert np.isclose(np.asarray(results['inverse_polarity']['1'][i]), np.asarray(result.polygons[i])).all()



from setuptools import setup, find_packages, Extension


setup(name='pygdsdesign',
      version='0.17.6',
      use_2to3=False,
      author='Étienne Dumur, Sacha Wos',
      author_email='etienne.dumur@gmail.com',
      maintainer='Étienne Dumur',
      maintainer_email='etienne.dumur@gmail.com',
      description='pygdsdesign provides some function to more efficiently create gds files.',
      long_description=open('README.md', encoding='utf8').read(),
      classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering'
      ],
      license='BSL',
      packages=find_packages(),
      ext_modules = [Extension("clipper", ["pygdsdesign/clipper/clipper.cpp"])],
      install_requires=[
          'numpy',
          'pytest',
          'scipy',
          'tqdm',
          'typing_extensions',
      ],
      )



from pygdsdesign import PolygonSet, GdsLibrary, Text, MicroStrip, Rectangle
import numpy as np


def test_constructor():
    # Check if we can initialize an empty polygon correctly
    result = PolygonSet()
    assert np.isclose(np.asarray(result.polygons), np.asarray([[[0, 0]]])).all()
    assert result.layers == [0]
    assert result.datatypes == [0]
    assert result.names == ['']
    assert result.colors == ['']

    # Check the Auto-completion of the arguments based on polygons
    result = PolygonSet([[(1,2),(3,4)],[(1,2),(3,4)]])
    assert np.isclose(np.asarray(result.polygons), np.asarray([[[1,2],[3,4]]])).all()
    assert result.layers == [0,0]
    assert result.datatypes == [0,0]
    assert result.names == ['','']
    assert result.colors == ['','']

    # Check the Auto-completion of the arguments based on layers
    result = PolygonSet([[(1,2),(3,4)]], [0,0,0])
    assert np.isclose(np.asarray(result.polygons), np.asarray([[[1,2],[3,4]]])).all()
    assert result.layers == [0,0,0]
    assert result.datatypes == [0,0,0]
    assert result.names == ['','', '']
    assert result.colors == ['','', '']


def test_polygons_addition():

    # Check a basic polygon addition
    result = PolygonSet([[(1,2),(3,4)]])
    temp = PolygonSet([[(1,2),(3,4)],[(1,2),(3,4)]])
    result = temp + result
    oracle = [[[1,2],[3,4]],[[1,2],[3,4]],[[(1,2),(3,4)]]]

    for i in range(len(oracle)):
        assert np.isclose(np.asarray(oracle[i]), np.asarray(result.polygons[i])).all()
    assert result.layers == [0,0,0]
    assert result.datatypes == [0,0,0]
    assert result.names == ['','','']
    assert result.colors == ['','','']



from .polygonSet import PolygonSet
from .polygons import *
from .operation import *
from .library import GdsLibrary
from .transmission_lines.cpw import CPW
from .transmission_lines.cpw_polar import CPWPolar
from .transmission_lines.microstrip import MicroStrip
from .transmission_lines.microstrip_polar import MicroStripPolar
from .functions import *
from .shapes import *
from .materials_color import colors





from typing import List, Literal, Optional, Dict, Tuple, Self
import numpy as np
import struct
import warnings
from itertools import chain
_mpone = np.array((-1.0, 1.0))
from io import IOBase


from pygdsdesign.typing_local import Coordinate, Coordinates


class PolygonSet():

    __slots__ = "layers", "datatypes", "polygons", "properties", "names", "colors"

    def __init__(self, polygons: Coordinates=None,
                       layers: List[int]=None,
                       datatypes: List[int]=None,
                       names: List[str]=None,
                       colors: List[str]=None,
                       **kwargs) -> None:

        # Getting rid of the mutability of python's default arguments
        if polygons is None:
            polygons = [np.array([[0., 0.]])]
        if layers is None:
            layers = [0]
        if datatypes is None:
            datatypes = [0]
        if names is None:
            names = ['']
        if colors is None:
            colors = ['']

        self.polygons: List[np.ndarray] = [np.asarray(p) for p in polygons]
        self.layers: List[int] = layers
        self.datatypes: List[int] = datatypes
        self.names: List[str]  = names
        self.colors: List[str] = colors
        self.properties = {}

        # to keep retrop-compatibility with "layer" old attribute
        # To allow wasy use of layer dictionnary
        if 'layer' in kwargs:
            self.layers = [kwargs['layer']]

        # Find the maximum array length
        max_length = max(len(self.polygons), len(self.layers), len(self.datatypes), len(self.names), len(self.colors))

        # Equalize the size of each argument
        for arg, last_element in [(self.polygons, self.polygons[-1]),
                                    (self.layers, self.layers[-1]),
                                    (self.datatypes, self.datatypes[-1]),
                                    (self.names, self.names[-1]),
                                    (self.colors, self.colors[-1])]:
            while len(arg) < max_length:
                arg.append(last_element)


    def __gt__(self, point: Coordinate) -> Self:
        """
        Add a vertice to the polygon in respect to the last vertice coordinate

        Parameters
        ----------
        point : tuple
            Point coordinate (x, y)
        """

        self.polygons[0] = np.vstack((self.polygons[0], self.polygons[0][-1]+np.asarray(point)))

        return self


    def __add__(self, polygon: Self) -> Self:
        """
        Add two polygons together

        Parameters
        -----------
        polygon : Polygon
            Polygon to be added
        """

        ### Carreful with mutability here!

        # Check if empty polygon -> we replace it by the non-empty one
        if len(self.polygons)==1:
            if (self.polygons[0]==np.array([[0,0]])).all():

                self.polygons = polygon.polygons
                self.layers = polygon.layers
                self.datatypes = polygon.datatypes
                self.names = polygon.names
                self.colors = polygon.colors

                return self

        self.polygons = self.polygons + polygon.polygons
        self.layers = self.layers + polygon.layers
        self.datatypes = self.datatypes + polygon.datatypes
        self.names = self.names + polygon.names
        self.colors = self.colors + polygon.colors

        return self


    def __str__(self) -> str:
        return (
            "PolygonSet ({} polygons, {} vertices, layers {}, datatypes {}, names {}, colors {})"
        ).format(
            len(self.polygons),
            sum([len(p) for p in self.polygons]),
            self.layers,
            self.datatypes,
            self.names,
            self.colors,
        )


    def get_area(self, by_spec=False) -> float | Dict[Tuple[int, int], float]:
        """
        Calculate the total area of this polygon set.

        Parameters
        ----------
        by_spec : bool
            If True, the return value is a dictionary with
            ``{(layer, datatype): area}``.

        Returns
        -------
        out : number, dictionary
            Area of this object.
        """
        if by_spec:
            path_area = {}
            for poly, key in zip(self.polygons, zip(self.layers, self.datatypes)):
                poly_area = 0
                for ii in range(1, len(poly) - 1):
                    poly_area += (poly[0][0] - poly[ii + 1][0]) * (
                        poly[ii][1] - poly[0][1]
                    ) - (poly[0][1] - poly[ii + 1][1]) * (poly[ii][0] - poly[0][0])
                if key in path_area:
                    path_area[key] += 0.5 * abs(poly_area)
                else:
                    path_area[key] = 0.5 * abs(poly_area)
        else:
            path_area = 0.
            for points in self.polygons:
                poly_area = 0
                point0 = points[0][0]
                point01 = points[0][1]
                for ii, point in enumerate(points[1:-1], start=1):
                    poly_area += (point0 - points[ii + 1][0]) * (point[1] - point01) - (point01 - points[ii + 1][1]) * (point[0] - point0)
                path_area += 0.5 * abs(poly_area)
        return path_area


    def get_center(self) -> Coordinate:
        """
        Return the center of the polygon.
        """

        a, b = self.get_bounding_box()

        return (a[0]+b[0])/2., (a[1]+b[1])/2.


    def get_bounding_box(self) -> Optional[Tuple[Coordinate, Coordinate]]:
        """
        Calculate the bounding box of the polygons.

        Returns
        -------
        out : Numpy array[2, 2] or None
            Bounding box of this polygon in the form [[x_min, y_min],
            [x_max, y_max]], or None if the polygon is empty.
        """
        if len(self.polygons) == 0:
            warnings.warn("[pygdsdesign] You try to 'get_bounding_box' on a empty polygonSet",stacklevel=4,)
            return None

        # Convert the polygons into a single 3D numpy array
        all_points = np.vstack(self.polygons)

        min_x = np.min(all_points[:, 0])
        min_y = np.min(all_points[:, 1])
        max_x = np.max(all_points[:, 0])
        max_y = np.max(all_points[:, 1])

        return np.array([[min_x, min_y], [max_x, max_y]])


    def get_size(self) -> Coordinate:
        """
        Return the size the polygon as length along (x, y).
        """

        a, b = self.get_bounding_box()
        return b[0]-a[0],  b[1]-a[1]


    def rotate(self, angle: float,
                     center: Coordinate=(0, 0)) -> Self:
        """
        Rotate this object.

        Parameters
        ----------
        angle : number
            The angle of rotation (in *radians*).
        center : array-like[2]
            Center point for the rotation.

        Returns
        -------
        out : `PolygonSet`
            This object.
        """
        ca = np.cos(angle)
        sa = np.sin(angle) * _mpone
        c0 = np.array(center)
        new_polys = []

        for points in self.polygons:
            pts = points - c0
            new_polys.append(pts * ca + pts[:, ::-1] * sa + c0)
        self.polygons = new_polys

        return self


    def remove_polygon_from_layer(self, layer: int) -> Self:
        """
        Remove all polygons of the given layer from the current PolygonSet.
        For instance, if the current PolygonSet contains Polygon with layer
        (1, 2, 3, 6, 9), using this method with `layer=3` as input parameter result
        in a PolygonSet containing Polygon with layer (1, 2, 6, 9).

        Args:
            layer : Layer number.
        """

        # We create a numpy mask of the single layer we want to keep
        ls = np.array(self.layers)
        mask = ls!=layer

        # We filter the array with the mask
        self.polygons  = list(np.array(self.polygons, dtype=object)[mask])
        self.layers    = list(ls[mask])
        self.datatypes = list(np.array(self.datatypes)[mask])
        self.colors    = list(np.array(self.colors)[mask])
        self.names     = list(np.array(self.names)[mask])

        return self


    def scale(self, scalex: float,
                    scaley: Optional[float]=None,
                    center: Tuple[float, float]=(0., 0.)) -> Self:
        """
        Scale this object.

        Parameters
        ----------
        scalex : number
            Scaling factor along the first axis.
        scaley : number or None
            Scaling factor along the second axis.  If None, same as
            `scalex`.
        center : array-like[2]
            Center point for the scaling operation.

        Returns
        -------
        out : `PolygonSet`
            This object.
        """
        c0 = np.array(center)
        s = scalex if scaley is None else np.array((scalex, scaley))
        self.polygons = [(points - c0) * s + c0 for points in self.polygons]
        return self


    def center(self) -> Self:
        """
        Centered the polygon to (0, 0).
        """

        x0, y0 = self.get_center()
        self.translate(-x0, -y0)

        return self


    def flip(self, axis: Literal['x', 'y']) -> Self:
        """
        Flip the polygon along its central x or y axis.

        Parameters
        ----------
        axis : float
            Central axis of the flip.
        """

        x0, y0 = self.get_center()
        self.translate(-x0, -y0)

        if axis.lower()=='x':
            self.polygons = [np.vstack((p[:, 0], p[:, 1]*-1)).T for p in self.polygons]
        elif axis.lower()=='y':
            self.polygons = [np.vstack((p[:, 0]*-1, p[:, 1])).T for p in self.polygons]
        else:
            raise ValueError('axis must be "x", "y"')

        self.translate(x0, y0)

        return self


    def translate(self, dx: float,
                        dy: float) -> 'PolygonSet':
        """
        Translate this polygon.

        Parameters
        ----------
        dx : number
            Distance to move in the x-direction.
        dy : number
            Distance to move in the y-direction.

        Returns
        -------
        out : `PolygonSet`
            This object.
        """
        vec = np.array((dx, dy))
        self.polygons = [points + vec for points in self.polygons]
        return self


    def to_gds(self, outfile: IOBase,
                     multiplier: float) -> None:
            """
            Convert this object to a series of GDSII elements.

            Parameters
            ----------
            outfile : open file
                Output to write the GDSII.
            multiplier : number
                A number that multiplies all dimensions written in the GDSII
                elements.
            """
            for ii, polygon in enumerate(self.polygons):
                if len(polygon) > 8190:
                    # warnings.warn(
                    #     "[pygdsdesign] Polygons with more than 8190 are not supported by the "
                    #     "official GDSII specification.  This GDSII file might not be "
                    #     "compatible with all readers.",
                    #     stacklevel=4,
                    # )
                    outfile.write(
                        struct.pack(
                            ">4Hh2Hh",
                            4,
                            0x0800,
                            6,
                            0x0D02,
                            self.layers[ii],
                            6,
                            0x0E02,
                            self.datatypes[ii],
                        )
                    )
                    xy = np.empty((polygon.shape[0] + 1, 2), dtype=">i4")
                    xy[:-1] = np.round(polygon * multiplier)
                    xy[-1] = xy[0]
                    i0 = 0
                    while i0 < xy.shape[0]:
                        i1 = min(i0 + 8190, xy.shape[0])
                        outfile.write(struct.pack(">2H", 4 + 8 * (i1 - i0), 0x1003))
                        outfile.write(xy[i0:i1].tobytes())
                        i0 = i1
                else:
                    outfile.write(
                        struct.pack(
                            ">4Hh2Hh2H",
                            4,
                            0x0800,
                            6,
                            0x0D02,
                            self.layers[ii],
                            6,
                            0x0E02,
                            self.datatypes[ii],
                            12 + 8 * len(polygon),
                            0x1003,
                        )
                    )
                    # TODO -> enlever le round() ?
                    xy = np.round(polygon * multiplier).astype(">i4")
                    outfile.write(xy.tobytes())
                    outfile.write(xy[0].tobytes())
                if self.properties is not None and len(self.properties) > 0:
                    size = 0
                    for attr, value in self.properties.items():
                        if len(value) % 2 != 0:
                            value = value + "\0"
                        outfile.write(
                            struct.pack(">5H", 6, 0x2B02, attr, 4 + len(value), 0x2C06)
                        )
                        outfile.write(value.encode("ascii"))
                        size += len(value) + 2
                    if size > 128:
                        warnings.warn(
                            "[pygdsdesign] Properties with size larger than 128 bytes are not "
                            "officially supported by the GDSII specification.  This file "
                            "might not be compatible with all readers.",
                            stacklevel=4,
                        )
                outfile.write(struct.pack(">2H", 4, 0x1100))


    def mirror(self, p1: Tuple[float, float],
                     p2: Tuple[float, float]=(0, 0)) -> Self:
        """
        Mirror the polygons over a line through points 1 and 2

        Parameters
        ----------
        p1 : array-like[2]
            first point defining the reflection line
        p2 : array-like[2]
            second point defining the reflection line

        Returns
        -------
        out : `PolygonSet`
            This object.
        """
        origin = np.array(p1)
        vec = np.array(p2) - origin
        vec_r = vec * (2 / np.inner(vec, vec))
        self.polygons = [
            np.outer(np.inner(points - origin, vec_r), vec) - points + 2 * origin
            for points in self.polygons
        ]
        return self


    def fracture(self, max_points: int=199,
                       precision: float=1e-3) -> Self:
        """
        Slice these polygons in the horizontal and vertical directions
        so that each resulting piece has at most `max_points`.  This
        operation occurs in place.

        Parameters
        ----------
        max_points : integer
            Maximal number of points in each resulting polygon (at least
            5 for the fracture to occur).
        precision : float
            Desired precision for rounding vertice coordinates.

        Returns
        -------
        out : `PolygonSet`
            This object.
        """
        if max_points > 4:
            ii = 0
            while ii < len(self.polygons):
                if len(self.polygons[ii]) > max_points:
                    pts0 = sorted(self.polygons[ii][:, 0])
                    pts1 = sorted(self.polygons[ii][:, 1])
                    ncuts = len(pts0) // max_points
                    if pts0[-1] - pts0[0] > pts1[-1] - pts1[0]:
                        # Vertical cuts
                        cuts = [
                            pts0[int(i * len(pts0) / (ncuts + 1.0) + 0.5)]
                            for i in range(1, ncuts + 1)
                        ]
                        chopped = clipper._chop(
                            self.polygons[ii], cuts, 0, 1 / precision
                        )
                    else:
                        # Horizontal cuts
                        cuts = [
                            pts1[int(i * len(pts1) / (ncuts + 1.0) + 0.5)]
                            for i in range(1, ncuts + 1)
                        ]
                        chopped = clipper._chop(
                            self.polygons[ii], cuts, 1, 1 / precision
                        )
                    self.polygons.pop(ii)
                    layer = self.layers.pop(ii)
                    datatype = self.datatypes.pop(ii)
                    color = self.colors.pop(ii)
                    name = self.names.pop(ii)
                    self.polygons.extend(
                        np.array(x) for x in chain.from_iterable(chopped)
                    )
                    npols = sum(len(c) for c in chopped)
                    self.layers.extend(layer for _ in range(npols))
                    self.datatypes.extend(datatype for _ in range(npols))
                    self.colors.extend(color for _ in range(npols))
                    self.names.extend(name for _ in range(npols))
                else:
                    ii += 1
        return self


    def fillet(self, radius: float|np.ndarray|list,
                     points_per_2pi: int=128) -> Self:
        """
        Round the corners of these polygons and fractures them into
        polygons with less vertices if necessary.

        Parameters
        ----------
        radius : number, array-like
            Radius of the corners.  If number: all corners filleted by
            that amount.  If array: specify fillet radii on a
            per-polygon basis (length must be equal to the number of
            polygons in this `PolygonSet`).  Each element in the array
            can be a number (all corners filleted by the same amount) or
            another array of numbers, one per polygon vertex.
            Alternatively, the array can be flattened to have one radius
            per `PolygonSet` vertex.
        points_per_2pi : integer
            Number of vertices used to approximate a full circle.  The
            number of vertices in each corner of the polygon will be the
            fraction of this number corresponding to the angle
            encompassed by that corner with respect to 2 pi.

        Returns
        -------
        out : `PolygonSet`
            This object.
        """
        two_pi = 2 * np.pi

        if np.isscalar(radius):
            radii = [[radius] * p.shape[0] for p in self.polygons]
        else:
            assert isinstance(radius, np.ndarray)
            if len(radius) == len(self.polygons):
                radii = []
                for r, p in zip(radius, self.polygons):
                    if np.isscalar(r):
                        radii.append([r] * p.shape[0])
                    else:
                        if len(r) != p.shape[0]:
                            raise ValueError(
                                "[GDSPY] Wrong length in fillet radius list.  "
                                "Found {} radii for polygon with {} vertices.".format(
                                    len(r), len(p.shape[0])
                                )
                            )
                        radii.append(r)
            else:
                total = sum(p.shape[0] for p in self.polygons)
                if len(radius) != total:
                    raise ValueError(
                        "[GDSPY] Wrong length in fillet radius list.  "
                        "Expected lengths are {} or {}; got {}.".format(
                            len(self.polygons), total, len(radius)
                        )
                    )
                radii = []
                n = 0
                for p in self.polygons:
                    radii.append(radius[n : n + p.shape[0]])
                    n += p.shape[0]

        for jj in range(len(self.polygons)):
            vec = self.polygons[jj].astype(float) - np.roll(self.polygons[jj], 1, 0)
            length = (vec[:, 0] ** 2 + vec[:, 1] ** 2) ** 0.5
            ii = np.flatnonzero(length)
            if len(ii) < len(length):
                self.polygons[jj] = np.array(self.polygons[jj][ii])
                radii[jj] = [radii[jj][i] for i in ii]
                vec = self.polygons[jj].astype(float) - np.roll(
                    self.polygons[jj], 1, 0
                )
                length = (vec[:, 0] ** 2 + vec[:, 1] ** 2) ** 0.5
            vec[:, 0] = vec[:, 0] / length
            vec[:, 1] = vec[:, 1] / length
            dvec = np.roll(vec, -1, 0) - vec
            norm = (dvec[:, 0] ** 2 + dvec[:, 1] ** 2) ** 0.5
            ii = np.flatnonzero(norm)
            dvec[ii, 0] = dvec[ii, 0] / norm[ii]
            dvec[ii, 1] = dvec[ii, 1] / norm[ii]
            dot = np.roll(vec, -1, 0) * vec
            theta = np.arccos(dot[:, 0] + dot[:, 1])
            ct = np.cos(theta * 0.5)
            tt = np.tan(theta * 0.5)

            new_points = []
            for ii in range(-1, len(self.polygons[jj]) - 1):
                if theta[ii] > 1e-6:
                    a0 = -vec[ii] * tt[ii] - dvec[ii] / ct[ii]
                    a0 = np.arctan2(a0[1], a0[0])
                    a1 = vec[ii + 1] * tt[ii] - dvec[ii] / ct[ii]
                    a1 = np.arctan2(a1[1], a1[0])
                    if a1 - a0 > np.pi:
                        a1 -= two_pi
                    elif a1 - a0 < -np.pi:
                        a1 += two_pi
                    n = max(
                        int(np.ceil(abs(a1 - a0) / two_pi * points_per_2pi) + 0.5), 2
                    )
                    a = np.linspace(a0, a1, n)
                    ll = radii[jj][ii] * tt[ii]
                    if ll > 0.49 * length[ii]:
                        r = 0.49 * length[ii] / tt[ii]
                        ll = 0.49 * length[ii]
                    else:
                        r = radii[jj][ii]
                    if ll > 0.49 * length[ii + 1]:
                        r = 0.49 * length[ii + 1] / tt[ii]
                    new_points.extend(
                        r * dvec[ii] / ct[ii]
                        + self.polygons[jj][ii]
                        + np.vstack((r * np.cos(a), r * np.sin(a))).transpose()
                    )
                else:
                    new_points.append(self.polygons[jj][ii])
            self.polygons[jj] = np.array(new_points)
        return self


    def change_layer(self, layer: Optional[int]=None,
                           datatype: Optional[int]=None,
                           color: Optional[str]=None,
                           name: Optional[str] = None) -> Self:
        """
        Change the layer properties of the polygons.
        Only given parameters will be updated in the polygon layer, all the other
        will stay unchanged.

        Args:
            layer : Layer number.
            datatype : Datatype number.
            color : Color hex code.
            name : Name of the layer
        """

        if layer is not None:
            for i in range(len(self.layers)):
                self.layers[i] = layer
        if datatype is not None:
            for i in range(len(self.datatypes)):
                self.datatypes[i] = datatype
        if color is not None:
            for i in range(len(self.colors)):
                self.colors[i] = color
        if name is not None:
            for i in range(len(self.names)):
                self.names[i] = name

        return self



import time

from pygdsdesign import GdsLibrary, PolygonSet, lateqs_logo

start = time.time()

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')


# Create empty polygon to which other polygon will be added
tot = PolygonSet()
tot += lateqs_logo(layer=2, datatype=1, name='logo')


# Add polygons to cell
cell.add(tot)

# Save GDS file
lib.export_gds('import_gds.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')
